import pytest

from autorag.nodes.passagereranker import Upr
from tests.autorag.nodes.passagereranker.test_passage_reranker_base import (
	base_reranker_test,
	queries_example,
	contents_example,
	ids_example,
	base_reranker_node_test,
	project_dir,
	previous_result,
)
from tests.delete_tests import is_github_action


@pytest.fixture
def upr_reranker():
	return Upr(project_dir)


@pytest.mark.skipif(is_github_action(), reason="Skipping this test on GitHub Actions")
def test_upr(upr_reranker):
	top_k = 1
	contents_result, id_result, score_result = upr_reranker._pure(
		queries_example, contents_example, ids_example, top_k
	)
	base_reranker_test(
		contents_result, id_result, score_result, top_k, descending=False
	)


@pytest.mark.skipif(is_github_action(), reason="Skipping this test on GitHub Actions")
def test_upr_node():
	top_k = 3
	result_df = Upr.run_evaluator(
		project_dir=project_dir, previous_result=previous_result, top_k=top_k
	)
	base_reranker_node_test(result_df, top_k, descending=False)



import os
import pathlib
import tempfile

import chromadb
import pandas as pd
import pytest
from llama_index.embeddings.openai import OpenAIEmbedding

from autorag.nodes.retrieval.vectordb import vectordb_ingest

root_dir = pathlib.PurePath(
	os.path.dirname(os.path.realpath(__file__))
).parent.parent.parent
project_dir = os.path.join(root_dir, "resources", "sample_project")
qa_data = pd.read_parquet(os.path.join(project_dir, "data", "qa.parquet"))
corpus_data = pd.read_parquet(os.path.join(project_dir, "data", "corpus.parquet"))
previous_result = qa_data.sample(5)

root_dir = pathlib.PurePath(
	os.path.dirname(os.path.realpath(__file__))
).parent.parent.parent
resource_path = os.path.join(root_dir, "resources")

embedding_model = OpenAIEmbedding()


@pytest.fixture
def ingested_vectordb_node():
	with tempfile.TemporaryDirectory(ignore_cleanup_errors=True) as node_chroma_path:
		node_db = chromadb.PersistentClient(path=node_chroma_path)
		node_collection = node_db.create_collection(
			name="openai", metadata={"hnsw:space": "cosine"}
		)

		node_test_corpus_path = os.path.join(
			resource_path, "sample_project", "data", "corpus.parquet"
		)
		sample_project_corpus_df = pd.read_parquet(path=node_test_corpus_path)

		vectordb_ingest(node_collection, sample_project_corpus_df, embedding_model)

		assert node_collection.count() == 30
		yield node_collection


def base_query_expansion_node_test(result_df):
	queries = result_df["queries"].tolist()
	assert len(queries) == 5
	assert all(isinstance(query, str) for query_list in queries for query in query_list)



from unittest.mock import patch

import cohere.base_client
import pytest
from cohere import RerankResponse, RerankResponseResultsItem

from autorag.nodes.passagereranker import CohereReranker
from tests.autorag.nodes.passagereranker.test_passage_reranker_base import (
	queries_example,
	contents_example,
	scores_example,
	ids_example,
	base_reranker_test,
	project_dir,
	previous_result,
	base_reranker_node_test,
)


async def mock_cohere_reranker(
	self, model, query, documents, top_n, **kwargs
) -> RerankResponse:
	if query == queries_example[0]:
		return RerankResponse(
			results=[
				RerankResponseResultsItem(index=1, relevance_score=0.8),
				RerankResponseResultsItem(index=2, relevance_score=0.2),
				RerankResponseResultsItem(index=0, relevance_score=0.1),
			][:top_n]
		)
	elif query == queries_example[1]:
		return RerankResponse(
			results=[
				RerankResponseResultsItem(index=1, relevance_score=0.8),
				RerankResponseResultsItem(index=0, relevance_score=0.2),
				RerankResponseResultsItem(index=2, relevance_score=0.1),
			][:top_n]
		)
	response_items = [
		RerankResponseResultsItem(index=i, relevance_score=0.1 * i)
		for i in range(len(documents) - 1, -1, -1)
	]
	return RerankResponse(results=response_items[:top_n])


@pytest.fixture
def cohere_reranker_instance():
	return CohereReranker(project_dir=project_dir, api_key="test")


@patch.object(cohere.base_client.AsyncBaseCohere, "rerank", mock_cohere_reranker)
def test_cohere_reranker(cohere_reranker_instance):
	top_k = 3
	contents_result, id_result, score_result = cohere_reranker_instance._pure(
		queries_example, contents_example, scores_example, ids_example, top_k
	)
	base_reranker_test(contents_result, id_result, score_result, top_k)


@patch.object(cohere.base_client.AsyncBaseCohere, "rerank", mock_cohere_reranker)
def test_cohere_reranker_batch_one(cohere_reranker_instance):
	top_k = 3
	batch = 1
	contents_result, id_result, score_result = cohere_reranker_instance._pure(
		queries_example,
		contents_example,
		scores_example,
		ids_example,
		top_k,
		batch=batch,
	)
	base_reranker_test(contents_result, id_result, score_result, top_k)


@patch.object(cohere.base_client.AsyncBaseCohere, "rerank", mock_cohere_reranker)
def test_cohere_node():
	top_k = 1
	result_df = CohereReranker.run_evaluator(
		project_dir=project_dir, previous_result=previous_result, top_k=top_k
	)
	base_reranker_node_test(result_df, top_k)



import pytest

from autorag.schema.module import Module
from autorag.support import get_support_modules


# Test cases for supported module types
@pytest.mark.parametrize(
	"module_type, expected_module",
	[
		("bm25", get_support_modules("bm25")),
		("vectordb", get_support_modules("vectordb")),
	],
)
def test_module_from_dict_supported(module_type, expected_module):
	module_dict = {
		"module_type": module_type,
		"param1": "value1",
		"param2": "value2",
	}
	module = Module.from_dict(module_dict)
	assert module.module_type == module_type
	assert module.module == expected_module
	assert module.module_param == {
		k: v for k, v in module_dict.items() if k != "module_type"
	}


# Test cases for unsupported module types
@pytest.mark.parametrize(
	"module_type",
	[
		"unsupported_module",
		"another_unsupported_module",
	],
)
def test_module_from_dict_unsupported(module_type):
	module_dict = {"module_type": module_type}
	with pytest.raises(KeyError):
		Module.from_dict(module_dict)



import os
import pathlib
from datetime import datetime

import pandas as pd
import pytest

from autorag.utils import (
	validate_qa_dataset,
	validate_corpus_dataset,
	cast_qa_dataset,
	cast_corpus_dataset,
	validate_qa_from_corpus_dataset,
)


@pytest.fixture
def qa_df():
	return pd.DataFrame(
		{
			"qid": ["id1", "id2", "id3"],
			"query": ["query1", "query2", "query3"],
			"retrieval_gt": [[["doc1", "doc3"], ["doc2"]], [[]], [[]]],
			"generation_gt": "answer1",
		}
	)


# Fixture for corpus dataset
@pytest.fixture
def corpus_df():
	return pd.DataFrame(
		{
			"doc_id": ["doc1", "doc2", "doc3", "doc4", "doc5"],
			"contents": ["content1", "content2", "content3", None, " \n  \t 	"],
			"metadata": [
				{"prev_id": None, "next_id": "doc2"},
				{"test_key": "test_value"},
				{"last_modified_datetime": datetime(2022, 12, 1, 3, 4, 5)},
				{"test_key": "test_value"},
				{"test_key": "test_value"},
			],
		}
	)


# Test validate_qa_dataset
def test_validate_qa_dataset(qa_df):
	# This should pass as the qa_df fixture contains the required columns
	validate_qa_dataset(qa_df)

	# This should fail as 'qid' column is missing
	with pytest.raises(AssertionError):
		invalid_df = qa_df.drop(columns=["qid"])
		validate_qa_dataset(invalid_df)


# Test validate_corpus_dataset
def test_validate_corpus_dataset(corpus_df):
	# This should pass as the corpus_df fixture contains the required columns
	validate_corpus_dataset(corpus_df)

	# This should fail as 'doc_id' column is missing
	with pytest.raises(AssertionError):
		invalid_df = corpus_df.drop(columns=["doc_id"])
		validate_corpus_dataset(invalid_df)


# Test cast_qa_dataset
def test_cast_qa_dataset(qa_df):
	# Cast the dataset and check for correct casting
	qa_df.drop(index=1, inplace=True)
	casted_df = cast_qa_dataset(qa_df)
	assert all(isinstance(x, list) for x in casted_df["retrieval_gt"])
	assert all(isinstance(x[0], list) for x in casted_df["retrieval_gt"])
	assert all(isinstance(x, list) for x in casted_df["generation_gt"])

	# Check for ValueError when 'retrieval_gt' has an incorrect type
	with pytest.raises(ValueError):
		invalid_df = qa_df.copy()
		invalid_df.at[0, "retrieval_gt"] = 123  # Invalid type for casting
		cast_qa_dataset(invalid_df)


# Test cast_corpus_dataset
def test_cast_corpus_dataset(corpus_df):
	# Cast the dataset and check for a datetime key in metadata
	casted_df = cast_corpus_dataset(corpus_df)
	assert len(casted_df) == 3
	assert all("last_modified_datetime" in x for x in casted_df["metadata"])
	assert all(
		isinstance(x["last_modified_datetime"], datetime) for x in casted_df["metadata"]
	)
	assert casted_df["metadata"].iloc[0]["prev_id"] is None
	assert casted_df["metadata"].iloc[0]["next_id"] == "doc2"
	assert casted_df["metadata"].iloc[1]["prev_id"] is None
	assert casted_df["metadata"].iloc[1]["next_id"] is None
	assert casted_df["metadata"].iloc[2]["next_id"] is None
	assert casted_df["metadata"].iloc[2]["last_modified_datetime"] == datetime(
		2022, 12, 1, 3, 4, 5
	)


def test_validate_qa_from_corpus_dataset(qa_df, corpus_df):
	validate_qa_from_corpus_dataset(qa_df, corpus_df)

	with pytest.raises(AssertionError) as excinfo:
		invalid_df = qa_df.copy()
		invalid_df.at[0, "retrieval_gt"] = [["answer1", "answer2"], ["answer3"]]
		validate_qa_from_corpus_dataset(invalid_df, corpus_df)
	assert "3 doc_ids in retrieval_gt do not exist in corpus_df." in str(excinfo.value)

	root_dir = pathlib.PurePath(
		os.path.dirname(os.path.realpath(__file__))
	).parent.parent
	project_dir = os.path.join(root_dir, "resources", "sample_project")
	qa_parquet = pd.read_parquet(os.path.join(project_dir, "data", "qa.parquet"))

	with pytest.raises(AssertionError) as excinfo:
		validate_qa_from_corpus_dataset(qa_parquet, corpus_df)
	assert "10 doc_ids in retrieval_gt do not exist in corpus_df." in str(excinfo.value)



import os
import pathlib
import tempfile

import pandas as pd
import pytest

from autorag.parser import Parser
from autorag.utils.util import load_summary_file

root_dir = pathlib.PurePath(os.path.dirname(os.path.realpath(__file__))).parent
resource_dir = os.path.join(root_dir, "resources")
result_dir = os.path.join(resource_dir, "test_results")
data_glob = os.path.join(resource_dir, "parse_data", "eng_text", "*")


@pytest.fixture
def parser():
	with tempfile.TemporaryDirectory(ignore_cleanup_errors=True) as project_dir:
		parser = Parser(data_path_glob=data_glob, project_dir=project_dir)
		yield parser


@pytest.fixture
def test_parser():
	with tempfile.TemporaryDirectory(ignore_cleanup_errors=True) as project_dir:
		parser = Parser(data_path_glob=data_glob, project_dir=project_dir)
		yield parser


def test_start_parsing(parser):
	parser.start_parsing(os.path.join(resource_dir, "simple_parse.yaml"))
	project_dir = parser.project_dir
	assert os.path.exists(os.path.join(project_dir, "0"))
	assert os.path.exists(os.path.join(project_dir, "0", "parse_config.yaml"))
	assert os.path.exists(os.path.join(project_dir, "0", "0.parquet"))
	pdfminer_result = pd.read_parquet(os.path.join(project_dir, "0", "0.parquet"))
	assert os.path.exists(os.path.join(project_dir, "0", "1.parquet"))
	pdfplumber_result = pd.read_parquet(os.path.join(project_dir, "0", "1.parquet"))

	expect_result_columns = ["texts", "path", "page", "last_modified_datetime"]
	assert all(
		[
			expect_column in pdfminer_result.columns
			for expect_column in expect_result_columns
		]
	)
	assert all(
		[
			expect_column in pdfplumber_result.columns
			for expect_column in expect_result_columns
		]
	)
	summary_df = load_summary_file(os.path.join(project_dir, "0", "summary.csv"))
	expect_summary_columns = [
		"filename",
		"module_name",
		"module_params",
		"execution_time",
	]
	assert all(
		[
			expect_column in summary_df.columns
			for expect_column in expect_summary_columns
		]
	)



import os
import tempfile

import pytest
from llama_index.core import Document
from llama_index.core.schema import TextNode, NodeRelationship, RelatedNodeInfo

from autorag.data.legacy.corpus import (
	llama_documents_to_parquet,
	llama_text_node_to_parquet,
)
from tests.autorag.data.legacy.corpus.test_base_corpus_legacy import validate_corpus


@pytest.fixture
def parquet_filepath():
	with tempfile.NamedTemporaryFile(suffix=".parquet", delete=False) as temp_file:
		yield temp_file.name
		temp_file.close()
		os.unlink(temp_file.name)


def test_llama_documents_to_parquet(parquet_filepath):
	documents = [
		Document(
			text="test text",
			metadata={"key": "value"},
		)
		for _ in range(5)
	]
	result_df = llama_documents_to_parquet(documents, parquet_filepath, upsert=True)
	validate_corpus(result_df, 5, parquet_filepath)


def test_llama_text_node_to_parquet(parquet_filepath):
	sample_text_node = TextNode(
		text="test text",
		metadata={"key": "value"},
		relationships={
			NodeRelationship.PREVIOUS: RelatedNodeInfo(node_id="0"),
			NodeRelationship.NEXT: RelatedNodeInfo(node_id="2"),
		},
	)
	text_nodes = [sample_text_node for _ in range(5)]
	result_df = llama_text_node_to_parquet(text_nodes, parquet_filepath, upsert=True)
	validate_corpus(result_df, 5, parquet_filepath)



import json
import os.path
from unittest.mock import patch

import pytest

import autorag
from autorag.data.parse.clova import (
	clova_ocr,
	pdf_to_images,
	generate_image_info,
	extract_text_from_fields,
	json_to_html_table,
)

from tests.autorag.data.parse.test_parse_base import (
	single_pdf_path_list,
	multiple_pdf_data_list,
	data_dir,
	korean_text_glob,
	eng_text_glob,
)


async def mock_clova_ocr_pure(
	image_data: bytes,
	image_name: str,
	url: str,
	api_key: str,
	table_detection: bool = False,
):
	return "Mocked OCR result", single_pdf_path_list[0], 1


def check_clova_result(texts, path, pages):
	assert texts == [
		"Mocked OCR result",
		"Mocked OCR result",
	]
	assert path == [
		single_pdf_path_list[0],
		single_pdf_path_list[0],
	]
	assert pages == [1, 1]


@pytest.fixture
def result_sample():
	result_path = os.path.join(data_dir, "clova_data", "result_sample.json")
	with open(result_path, "r", encoding="utf-8") as file:
		result_sample = json.load(file)
		yield result_sample


@pytest.fixture
def expect_text():
	text_path = os.path.join(data_dir, "clova_data", "result_text.txt")
	with open(text_path, "r", encoding="utf-8") as file:
		expect_text = file.read()
		expect_text = expect_text.split("\n\\n\\n\\n")[0]
		yield expect_text


@pytest.fixture
def expect_table():
	table_path = os.path.join(data_dir, "clova_data", "result_table.txt")
	with open(table_path, "r", encoding="utf-8") as file:
		expect_table = file.read()
		expect_table = expect_table.split("\n\\n\\n\\n")[0]
		yield expect_table


@patch.object(autorag.data.parse.clova, "clova_ocr_pure", mock_clova_ocr_pure)
def test_clova_ocr_single_pdf():
	clova_ocr_original = clova_ocr.__wrapped__
	texts, path, pages = clova_ocr_original(
		single_pdf_path_list, url="mock_url", api_key="mock_api_key"
	)
	check_clova_result(texts, path, pages)


@patch.object(autorag.data.parse.clova, "clova_ocr_pure", mock_clova_ocr_pure)
def test_clova_ocr_single_pdf_node():
	result_df = clova_ocr(korean_text_glob, url="mock_url", api_key="mock_api_key")
	check_clova_result(
		result_df["texts"].tolist(),
		result_df["path"].tolist(),
		result_df["page"].tolist(),
	)


@patch.object(autorag.data.parse.clova, "clova_ocr_pure", mock_clova_ocr_pure)
def test_clova_ocr_multiple_pdf():
	clova_ocr_original = clova_ocr.__wrapped__
	texts, path, pages = clova_ocr_original(
		multiple_pdf_data_list, url="mock_url", api_key="mock_api_key"
	)
	check_clova_result(texts, path, pages)


@patch.object(autorag.data.parse.clova, "clova_ocr_pure", mock_clova_ocr_pure)
def test_clova_ocr_multiple_pdf_node():
	result_df = clova_ocr(eng_text_glob, url="mock_url", api_key="mock_api_key")
	check_clova_result(
		result_df["texts"].tolist(),
		result_df["path"].tolist(),
		result_df["page"].tolist(),
	)


def test_pdf_to_images():
	data_list = pdf_to_images(multiple_pdf_data_list[0])
	assert isinstance(data_list, list)
	assert isinstance(data_list[0], bytes)


def test_generate_image_info():
	names = generate_image_info(single_pdf_path_list[0], 2)
	assert names == [
		{"pdf_path": single_pdf_path_list[0], "pdf_page": 1},
		{"pdf_path": single_pdf_path_list[0], "pdf_page": 2},
	]


def test_extract_text_from_fields(result_sample, expect_text):
	text = extract_text_from_fields(result_sample["images"][0]["fields"])
	assert text == expect_text


def test_json_to_html_table(result_sample, expect_table):
	table = json_to_html_table(result_sample["images"][0]["tables"][0]["cells"])
	assert table == expect_table



import pytest

from autorag.nodes.passagereranker import FlagEmbeddingLLMReranker
from tests.autorag.nodes.passagereranker.test_passage_reranker_base import (
	queries_example,
	contents_example,
	ids_example,
	base_reranker_test,
	project_dir,
	previous_result,
	base_reranker_node_test,
)
from tests.delete_tests import is_github_action


@pytest.fixture
def flag_embedding_llm_reranker():
	return FlagEmbeddingLLMReranker(
		project_dir, "BAAI/bge-reranker-v2-gemma", use_fp16=False
	)


@pytest.mark.skip(reason="This test is too slow on the local macbook pro")
@pytest.mark.skipif(is_github_action(), reason="Skipping this test on GitHub Actions")
def test_flag_embedding_llm_reranker(flag_embedding_llm_reranker):
	top_k = 3
	contents_result, id_result, score_result = flag_embedding_llm_reranker._pure(
		queries_example,
		contents_example,
		ids_example,
		top_k,
	)
	base_reranker_test(contents_result, id_result, score_result, top_k)


@pytest.mark.skip(reason="This test is too slow on the local macbook pro")
@pytest.mark.skipif(is_github_action(), reason="Skipping this test on GitHub Actions")
def test_flag_embedding_llm_reranker_batch_one(flag_embedding_llm_reranker):
	top_k = 3
	batch = 1
	contents_result, id_result, score_result = flag_embedding_llm_reranker._pure(
		queries_example,
		contents_example,
		ids_example,
		top_k,
		batch,
	)
	base_reranker_test(contents_result, id_result, score_result, top_k)


@pytest.mark.skip(reason="This test is too slow on the local macbook pro")
@pytest.mark.skipif(is_github_action(), reason="Skipping this test on GitHub Actions")
def test_flag_embedding_llm_reranker_batch(flag_embedding_llm_reranker):
	top_k = 3
	batch = 2
	contents_result, id_result, score_result = flag_embedding_llm_reranker._pure(
		queries_example,
		contents_example,
		ids_example,
		top_k,
		batch,
	)
	base_reranker_test(contents_result, id_result, score_result, top_k)


@pytest.mark.skip(reason="This test is too slow on the local macbook pro")
@pytest.mark.skipif(is_github_action(), reason="Skipping this test on GitHub Actions")
def test_flag_embedding_llm_reranker_node():
	top_k = 1
	result_df = FlagEmbeddingLLMReranker.run_evaluator(
		project_dir=project_dir, previous_result=previous_result, top_k=top_k
	)
	base_reranker_node_test(result_df, top_k)



from unittest.mock import patch

import aiohttp
import pytest
from aioresponses import aioresponses

import autorag
from autorag.nodes.passagereranker import JinaReranker
from autorag.nodes.passagereranker.jina import jina_reranker_pure, JINA_API_URL
from tests.autorag.nodes.passagereranker.test_passage_reranker_base import (
	queries_example,
	contents_example,
	ids_example,
	base_reranker_test,
	project_dir,
	previous_result,
	base_reranker_node_test,
)


@pytest.mark.asyncio()
async def test_jina_reranker_pure():
	with aioresponses() as m:
		mock_response = {
			"results": [
				{"index": 1, "relevance_score": 0.9},
				{"index": 0, "relevance_score": 0.2},
			]
		}
		m.post(JINA_API_URL, payload=mock_response)
		session = aiohttp.ClientSession()
		session.headers.update(
			{"Authorization": "Bearer mock_api_key", "Accept-Encoding": "identity"}
		)
		content_result, id_result, score_result = await jina_reranker_pure(
			session,
			queries_example[0],
			contents_example[0],
			ids_example[0],
			top_k=2,
		)
		assert len(content_result) == 2
		assert len(id_result) == 2
		assert len(score_result) == 2

		assert all([res in contents_example[0] for res in content_result])
		assert all([res in ids_example[0] for res in id_result])

		# check if the scores are sorted
		assert score_result[0] >= score_result[1]


async def mock_jina_reranker_pure(session, query, contents, ids, top_k, **kwargs):
	if query == queries_example[0]:
		return (
			[contents[1], contents[2], contents[0]][:top_k],
			[ids[1], ids[2], ids[0]][:top_k],
			[0.8, 0.2, 0.1][:top_k],
		)
	elif query == queries_example[1]:
		return (
			[contents[1], contents[0], contents[2]][:top_k],
			[ids[1], ids[0], ids[2]][:top_k],
			[0.8, 0.2, 0.1][:top_k],
		)
	else:
		raise ValueError(f"Unexpected query: {query}")


@pytest.fixture
def jina_reranker_instance():
	return JinaReranker(project_dir, "mock_api_key")


@patch.object(
	autorag.nodes.passagereranker.jina, "jina_reranker_pure", mock_jina_reranker_pure
)
def test_jina_reranker(jina_reranker_instance):
	top_k = 3
	contents_result, id_result, score_result = jina_reranker_instance._pure(
		queries_example, contents_example, ids_example, top_k
	)
	base_reranker_test(contents_result, id_result, score_result, top_k)


@patch.object(
	autorag.nodes.passagereranker.jina, "jina_reranker_pure", mock_jina_reranker_pure
)
def test_jina_reranker_batch_one(jina_reranker_instance):
	top_k = 3
	batch = 1
	contents_result, id_result, score_result = jina_reranker_instance._pure(
		queries_example,
		contents_example,
		ids_example,
		top_k,
		batch=batch,
	)
	base_reranker_test(contents_result, id_result, score_result, top_k)


@patch.object(
	autorag.nodes.passagereranker.jina, "jina_reranker_pure", mock_jina_reranker_pure
)
def test_jina_reranker_node():
	top_k = 1
	result_df = JinaReranker.run_evaluator(
		project_dir=project_dir, previous_result=previous_result, top_k=top_k
	)
	base_reranker_node_test(result_df, top_k)



from unittest.mock import patch

import pytest
from llama_index.embeddings.openai import OpenAIEmbedding

from autorag.nodes.passageaugmenter import PrevNextPassageAugmenter

from tests.autorag.nodes.passageaugmenter.test_base_passage_augmenter import (
	ids_list,
	project_dir,
	previous_result,
	corpus_data,
	doc_id_list,
)
from tests.mock import mock_get_text_embedding_batch


@pytest.fixture
def prev_next_augmenter_instance():
	return PrevNextPassageAugmenter(project_dir=project_dir)


def test_prev_next_augmenter_next(prev_next_augmenter_instance):
	results = prev_next_augmenter_instance._pure(ids_list, num_passages=1, mode="next")
	assert results == [
		[doc_id_list[1], doc_id_list[2]],
		[doc_id_list[3], doc_id_list[4]],
		[doc_id_list[0], doc_id_list[1], doc_id_list[29]],
	]


def test_prev_next_augmenter_prev(prev_next_augmenter_instance):
	results = prev_next_augmenter_instance._pure(ids_list, num_passages=1, mode="prev")
	assert results == [
		[doc_id_list[0], doc_id_list[1]],
		[doc_id_list[2], doc_id_list[3]],
		[doc_id_list[0], doc_id_list[28], doc_id_list[29]],
	]


def test_prev_next_augmenter_both(prev_next_augmenter_instance):
	results = prev_next_augmenter_instance._pure(ids_list, num_passages=1, mode="both")
	assert results == [
		[doc_id_list[0], doc_id_list[1], doc_id_list[2]],
		[doc_id_list[2], doc_id_list[3], doc_id_list[4]],
		[doc_id_list[0], doc_id_list[1], doc_id_list[28], doc_id_list[29]],
	]


def test_prev_next_augmenter_multi_passages(prev_next_augmenter_instance):
	results = prev_next_augmenter_instance._pure(ids_list, num_passages=3, mode="prev")
	assert results == [
		[doc_id_list[0], doc_id_list[1]],
		[doc_id_list[0], doc_id_list[1], doc_id_list[2], doc_id_list[3]],
		[
			doc_id_list[0],
			doc_id_list[26],
			doc_id_list[27],
			doc_id_list[28],
			doc_id_list[29],
		],
	]


@patch.object(
	OpenAIEmbedding,
	"get_text_embedding_batch",
	mock_get_text_embedding_batch,
)
def test_prev_next_augmenter_node():
	result_df = PrevNextPassageAugmenter.run_evaluator(
		project_dir=project_dir, previous_result=previous_result, mode="next", top_k=2
	)
	contents = result_df["retrieved_contents"].tolist()
	ids = result_df["retrieved_ids"].tolist()
	scores = result_df["retrieve_scores"].tolist()
	assert len(contents) == len(ids) == len(scores) == 2
	assert len(contents[0]) == len(ids[0]) == len(scores[0]) == 2
	for content_list, id_list, score_list in zip(contents, ids, scores):
		for i, (content, _id, score) in enumerate(
			zip(content_list, id_list, score_list)
		):
			assert isinstance(content, str)
			assert isinstance(_id, str)
			assert isinstance(score, float)
			assert _id in corpus_data["doc_id"].tolist()
			assert (
				content
				== corpus_data[corpus_data["doc_id"] == _id]["contents"].values[0]
			)
			if i >= 1:
				assert score_list[i - 1] >= score_list[i]



import pytest

from autorag.data import chunk_modules
from autorag.data.chunk import langchain_chunk

from tests.autorag.data.chunk.test_chunk_base import (
	base_texts,
	parsed_result,
	check_chunk_result,
	base_metadata,
	character_expect_texts,
	expect_texts,
	expect_character_idx,
	expect_character_path,
	check_chunk_result_node,
)


@pytest.fixture
def chunk_instance():
	chunk_method = "character"
	kwargs = {"separator": ". ", "chunk_size": 30, "chunk_overlap": 0}
	chunk_instance = chunk_modules[chunk_method](**kwargs)
	yield chunk_instance


def test_langchain_chunk(chunk_instance):
	langchain_chunk_original = langchain_chunk.__wrapped__
	doc_id, contents, path, start_end_idx, metadata = langchain_chunk_original(
		base_texts, chunk_instance, metadata_list=base_metadata
	)
	check_chunk_result(doc_id, contents, path, start_end_idx, metadata)
	assert len(contents) == 7
	assert contents == character_expect_texts
	assert path == expect_character_path
	assert start_end_idx == expect_character_idx


def test_langchain_chunk_node(chunk_instance):
	result_df = langchain_chunk(parsed_result, chunk_method="character")
	check_chunk_result_node(result_df)
	assert len(result_df["doc_id"].tolist()) == 9


def test_langchain_chunk_file_name_ko(chunk_instance):
	langchain_chunk_original = langchain_chunk.__wrapped__
	doc_id, contents, path, start_end_idx, metadata = langchain_chunk_original(
		base_texts,
		chunk_instance,
		file_name_language="ko",
		metadata_list=base_metadata,
	)
	check_chunk_result(doc_id, contents, path, start_end_idx, metadata)
	assert len(contents) == 7
	assert contents == expect_texts["character_ko"]
	assert path == expect_character_path
	assert start_end_idx == expect_character_idx


def test_langchain_chunk_file_name_ko_node(chunk_instance):
	result_df = langchain_chunk(
		parsed_result, chunk_method="character", add_file_name="ko"
	)
	check_chunk_result_node(result_df)
	assert len(result_df["doc_id"].tolist()) == 9


def test_langchain_chunk_file_name_eng(chunk_instance):
	langchain_chunk_original = langchain_chunk.__wrapped__
	doc_id, contents, path, start_end_idx, metadata = langchain_chunk_original(
		base_texts,
		chunk_instance,
		file_name_language="en",
		metadata_list=base_metadata,
	)
	check_chunk_result(doc_id, contents, path, start_end_idx, metadata)
	assert len(contents) == 7
	assert contents == expect_texts["character_eng"]
	assert path == expect_character_path
	assert start_end_idx == expect_character_idx


def test_langchain_chunk_file_name_eng_node():
	result_df = langchain_chunk(
		parsed_result, chunk_method="character", add_file_name="en"
	)
	check_chunk_result_node(result_df)
	assert len(result_df["doc_id"].tolist()) == 9



import functools
from typing import List

import numpy as np

from autorag.schema.metricinput import MetricInput
from autorag.utils.util import convert_inputs_to_list


def calculate_cosine_similarity(a, b):
    dot_product = np.dot(a, b)
    norm_a = np.linalg.norm(a)
    norm_b = np.linalg.norm(b)
    similarity = dot_product / (norm_a * norm_b)
    return similarity


def autorag_metric(fields_to_check: List[str]):
    def decorator_autorag_metric(func):
        @functools.wraps(func)
        @convert_inputs_to_list
        def wrapper(
                metric_inputs: List[MetricInput], **kwargs
        ) -> List[float]:
            """
            :param metric_inputs: A list MetricInput schema for AutoRAG metric.
            :param kwargs: The additional arguments for metric function.
            :return: A list of computed metric scores.
            """
            results = []
            for metric_input in metric_inputs:
                if metric_input.is_fields_notnone(fields_to_check=fields_to_check):
                    results.append(func(metric_input, **kwargs))
                else:
                    results.append(None)
            return results

        return wrapper

    return decorator_autorag_metric


def autorag_metric_loop(fields_to_check: List[str]):
    def decorator_autorag_generation_metric(func):
        @functools.wraps(func)
        @convert_inputs_to_list
        def wrapper(
                metric_inputs: List[MetricInput], **kwargs
        ) -> List[float]:
            """
            :param metric_inputs: A list MetricInput schema for AutoRAG metric.
            :param kwargs: The additional arguments for metric function.
            :return: A list of computed metric scores.
            """
            bool_list = [metric_input.is_fields_notnone(fields_to_check=fields_to_check) for metric_input in
                         metric_inputs]
            valid_inputs = [metric_input for metric_input, is_valid in zip(metric_inputs, bool_list) if is_valid]

            results = [None] * len(metric_inputs)
            if valid_inputs:
                processed_valid = func(valid_inputs, **kwargs)

                valid_index = 0
                for i, is_valid in enumerate(bool_list):
                    if is_valid:
                        results[i] = processed_valid[valid_index]
                        valid_index += 1

            return results

        return wrapper

    return decorator_autorag_generation_metric



from dataclasses import dataclass
from typing import Optional, List, Dict, Callable, Any, Union

import numpy as np
import pandas as pd


@dataclass
class MetricInput:
    query: Optional[str] = None
    queries: Optional[List[str]] = None
    retrieval_gt_contents: Optional[List[List[str]]] = None
    retrieved_contents: Optional[List[str]] = None
    retrieval_gt: Optional[List[List[str]]] = None
    retrieved_ids: Optional[List[str]] = None
    prompt: Optional[str] = None
    generated_texts: Optional[str] = None
    generation_gt: Optional[List[str]] = None
    generated_log_probs: Optional[List[float]] = None

    def is_fields_notnone(self, fields_to_check: List[str]) -> bool:
        for field in fields_to_check:
            actual_value = getattr(self, field)

            if actual_value is None:
                return False

            try:
                if not type_checks.get(type(actual_value), lambda _: False)(actual_value):
                    return False
            except Exception:
                return False

        return True

    @classmethod
    def from_dataframe(cls, qa_data: pd.DataFrame) -> List['MetricInput']:
        """
        Convert a pandas DataFrame into a list of MetricInput instances.
        qa_data: pd.DataFrame: qa_data DataFrame containing metric data.

        :returns: List[MetricInput]: List of MetricInput objects created from DataFrame rows.
        """
        instances = []

        for _, row in qa_data.iterrows():
            instance = cls()

            for attr_name in cls.__annotations__:
                if attr_name in row:
                    value = row[attr_name]

                    if isinstance(value, str):
                        setattr(instance, attr_name, value.strip() if value.strip() != '' else None)
                    elif isinstance(value, list):
                        setattr(instance, attr_name, value if len(value) > 0 else None)
                    else:
                        setattr(instance, attr_name, value)

            instances.append(instance)

        return instances

    @staticmethod
    def _check_list(lst_or_arr: Union[List[Any], np.ndarray]) -> bool:
        if isinstance(lst_or_arr, np.ndarray):
            lst_or_arr = lst_or_arr.flatten().tolist()

        if len(lst_or_arr) == 0:
            return False

        for item in lst_or_arr:
            if item is None:
                return False

            item_type = type(item)

            if item_type in type_checks:
                if not type_checks[item_type](item):
                    return False
            else:
                return False

        return True


type_checks: Dict[type, Callable[[Any], bool]] = {
    str: lambda x: len(x.strip()) > 0,
    list: MetricInput._check_list,
    np.ndarray: MetricInput._check_list,
    int: lambda _: True,
    float: lambda _: True,
}



import os
import pathlib

import pandas as pd
import pytest

from autorag import dashboard
from autorag.dashboard import get_metric_values, make_trial_summary_md

root_dir = pathlib.PurePath(os.path.dirname(os.path.realpath(__file__))).parent
sample_project_dir = os.path.join(root_dir, "resources", "result_project")
sample_trial_dir = os.path.join(sample_project_dir, "0")


@pytest.fixture
def retrieval_summary_df():
	return pd.read_csv(
		os.path.join(sample_trial_dir, "retrieve_node_line", "retrieval", "summary.csv")
	)


def test_get_metric_values(retrieval_summary_df):
	result_dict = get_metric_values(retrieval_summary_df)
	assert len(result_dict.keys()) == 3
	assert set(list(result_dict.keys())) == {
		"retrieval_f1",
		"retrieval_recall",
		"retrieval_precision",
	}
	assert result_dict["retrieval_recall"] == 1.0
	assert result_dict["retrieval_precision"] == 0.1


def test_make_trial_summary_md():
	md_text = make_trial_summary_md(sample_trial_dir)
	assert bool(md_text)


@pytest.mark.skip(
	reason="Can't stop this test on the github action or pytest cli setup"
)
def test_dashboard_run():
	dashboard.run(sample_trial_dir)



from .generation import bleu, meteor, rouge, sem_score, g_eval, bert_score
from .retrieval import (
	retrieval_f1,
	retrieval_recall,
	retrieval_precision,
	retrieval_mrr,
	retrieval_ndcg,
	retrieval_map,
)
from .retrieval_contents import (
	retrieval_token_f1,
	retrieval_token_precision,
	retrieval_token_recall,
)



import pytest

from autorag.nodes.queryexpansion import MultiQueryExpansion
from tests.autorag.nodes.queryexpansion.test_query_expansion_base import (
	project_dir,
	previous_result,
	base_query_expansion_node_test,
)

sample_query = [
	"What are the potential benefits and drawbacks of implementing artificial intelligence in healthcare?",
	"How many members are in the group Aespa?",
]


@pytest.fixture
def multi_query_expansion_instance():
	return MultiQueryExpansion(
		project_dir=project_dir, generator_module_type="llama_index_llm", llm="mock"
	)


def test_multi_query_expansion(multi_query_expansion_instance):
	result = multi_query_expansion_instance._pure(
		sample_query,
		prompt="Generate Multi Query Expansion : {query}",
	)
	assert len(result[0]) > 1
	assert len(result) == 2
	assert isinstance(result[0][0], str)


def test_multi_query_expansion_node():
	generator_dict = {"generator_module_type": "llama_index_llm", "llm": "mock"}
	result_df = MultiQueryExpansion.run_evaluator(
		project_dir=project_dir, previous_result=previous_result, **generator_dict
	)
	base_query_expansion_node_test(result_df)



import pytest

from autorag.nodes.passagereranker import TimeReranker

from tests.autorag.nodes.passagereranker.test_passage_reranker_base import (
	contents_example,
	scores_example,
	ids_example,
	base_reranker_test,
	previous_result,
	base_reranker_node_test,
	time_list,
	project_dir_with_corpus,
)


@pytest.fixture
def time_reranker_instance(project_dir_with_corpus):
	return TimeReranker(project_dir_with_corpus)


def test_time_reranker(time_reranker_instance):
	top_k = 2

	contents_result, id_result, score_result = time_reranker_instance._pure(
		contents_example, scores_example, ids_example, top_k, time_list
	)
	assert id_result[0] == [ids_example[0][1], ids_example[0][2]]
	assert id_result[1] == [ids_example[1][1], ids_example[1][0]]
	assert contents_result[0] == [contents_example[0][1], contents_example[0][2]]
	assert contents_result[1] == [contents_example[1][1], contents_example[1][0]]
	assert score_result[0] == [0.8, 0.1]
	assert score_result[1] == [0.2, 0.1]
	base_reranker_test(contents_result, id_result, score_result, top_k)


def test_time_reranker_node(project_dir_with_corpus):
	top_k = 1
	result_df = TimeReranker.run_evaluator(
		project_dir=project_dir_with_corpus,
		previous_result=previous_result,
		top_k=top_k,
	)
	base_reranker_node_test(result_df, top_k)



from llama_index.core.llms import MockLLM

from autorag.data.qa.generation_gt.llama_index_gen_gt import (
	make_concise_gen_gt,
	make_basic_gen_gt,
)
from autorag.data.qa.schema import QA
from tests.autorag.data.qa.generation_gt.base_test_generation_gt import (
	qa_df,
	check_generation_gt,
)

llm = MockLLM()


def test_make_concise_gen_gt():
	qa = QA(qa_df)
	result_qa = qa.batch_apply(make_concise_gen_gt, llm=llm)
	check_generation_gt(result_qa)


def test_make_basic_gen_gt():
	qa = QA(qa_df)
	result_qa = qa.batch_apply(make_basic_gen_gt, llm=llm)
	check_generation_gt(result_qa)


def test_make_basic_gen_gt_ko():
	qa = QA(qa_df)
	result_qa = qa.batch_apply(make_basic_gen_gt, llm=llm, lang="ko")
	check_generation_gt(result_qa)


def test_make_multiple_gen_gt():
	qa = QA(qa_df)
	result_qa = qa.batch_apply(make_basic_gen_gt, llm=llm).batch_apply(
		make_concise_gen_gt, llm=llm
	)
	check_generation_gt(result_qa)
	assert all(len(x) == 2 for x in result_qa.data["generation_gt"].tolist())



import os
import tempfile

import pandas as pd

from autorag.data.qa.schema import Raw, Corpus, QA
from tests.autorag.data.qa.test_data_creation_piepline import initial_raw


def test_raw_add():
	raw_1 = Raw(
		pd.DataFrame(
			{
				"texts": ["hello", "world"],
				"path": ["path1", "path1"],
				"page": [1, 2],
				"last_modified_datetime": ["2021-08-01", "2021-08-02"],
			}
		)
	)
	raw_2 = Raw(
		pd.DataFrame(
			{
				"texts": ["foo", "bar"],
				"path": ["path2", "path2"],
				"page": [3, 4],
				"last_modified_datetime": ["2021-08-03", "2021-08-04"],
			}
		)
	)
	expected_dataframe = pd.DataFrame(
		{
			"texts": ["hello", "world", "foo", "bar"],
			"path": ["path1", "path1", "path2", "path2"],
			"page": [1, 2, 3, 4],
			"last_modified_datetime": [
				"2021-08-01",
				"2021-08-02",
				"2021-08-03",
				"2021-08-04",
			],
		}
	)
	pd.testing.assert_frame_equal((raw_1 + raw_2).data, expected_dataframe)


def test_raw_chunk():
	corpus = initial_raw.chunk(
		"llama_index_chunk", chunk_method="token", chunk_size=128, chunk_overlap=5
	)
	assert isinstance(corpus, Corpus)
	pd.testing.assert_frame_equal(corpus.linked_raw.data, initial_raw.data)
	assert set(corpus.data.columns) == {
		"doc_id",
		"contents",
		"path",
		"start_end_idx",
		"metadata",
	}
	assert corpus.data["doc_id"].nunique() == len(corpus.data)
	assert all(
		origin_path in initial_raw.data["path"].tolist()
		for origin_path in corpus.data["path"].tolist()
	)


def test_update_corpus():
	raw = Raw(
		pd.DataFrame(
			{
				"texts": ["hello", "world", "jax"],
				"path": ["path1", "path1", "path2"],
				"page": [1, 2, -1],
				"last_modified_datetime": [
					"2021-08-01",
					"2021-08-02",
					"2021-08-03",
				],
			}
		)
	)
	original_corpus = Corpus(
		pd.DataFrame(
			{
				"doc_id": ["id1", "id2", "id3", "id4", "id5", "id6"],
				"contents": ["hello", "world", "foo", "bar", "baz", "jax"],
				"path": ["path1", "path1", "path1", "path1", "path2", "path2"],
				"start_end_idx": [
					(0, 120),
					(90, 200),
					(0, 40),
					(35, 75),
					(0, 100),
					(150, 200),
				],
				"metadata": [
					{"page": 1, "last_modified_datetime": "2021-08-01"},
					{"page": 1, "last_modified_datetime": "2021-08-01"},
					{"page": 2, "last_modified_datetime": "2021-08-02"},
					{"page": 2, "last_modified_datetime": "2021-08-02"},
					{"last_modified_datetime": "2021-08-01"},
					{"last_modified_datetime": "2021-08-01"},
				],
			}
		),
		raw,
	)

	qa = QA(
		pd.DataFrame(
			{
				"qid": ["qid1", "qid2", "qid3", "qid4"],
				"query": ["hello", "world", "foo", "bar"],
				"retrieval_gt": [
					[["id1"]],
					[["id1"], ["id2"]],
					[["id3", "id4"]],
					[["id6", "id2"], ["id5"]],
				],
				"generation_gt": ["world", "foo", "bar", "jax"],
			}
		),
		original_corpus,
	)

	new_corpus = Corpus(
		pd.DataFrame(
			{
				"doc_id": [
					"new_id1",
					"new_id2",
					"new_id3",
					"new_id4",
					"new_id5",
					"new_id6",
				],
				"contents": ["hello", "world", "foo", "bar", "baz", "jax"],
				"path": ["path1", "path1", "path1", "path1", "path2", "path2"],
				"start_end_idx": [
					(0, 80),
					(80, 150),
					(15, 50),
					(50, 80),
					(0, 200),
					(201, 400),
				],
				"metadata": [
					{"page": 1, "last_modified_datetime": "2021-08-01"},
					{"page": 1, "last_modified_datetime": "2021-08-01"},
					{"page": 2, "last_modified_datetime": "2021-08-02"},
					{"page": 2, "last_modified_datetime": "2021-08-02"},
					{"last_modified_datetime": "2021-08-01"},
					{"last_modified_datetime": "2021-08-01"},
				],
			}
		),
		raw,
	)

	new_qa = qa.update_corpus(new_corpus)

	expected_dataframe = pd.DataFrame(
		{
			"qid": ["qid1", "qid2", "qid3", "qid4"],
			"retrieval_gt": [
				[["new_id1", "new_id2"]],
				[["new_id1", "new_id2"], ["new_id2"]],
				[["new_id3", "new_id3", "new_id4"]],
				[["new_id5", "new_id2"], ["new_id5"]],
			],
		}
	)
	pd.testing.assert_frame_equal(
		new_qa.data[["qid", "retrieval_gt"]], expected_dataframe
	)
	with tempfile.NamedTemporaryFile(suffix=".parquet", delete=False) as qa_path:
		with tempfile.NamedTemporaryFile(
			suffix=".parquet", delete=False
		) as corpus_path:
			new_qa.to_parquet(qa_path.name, corpus_path.name)
			loaded_qa = pd.read_parquet(qa_path.name, engine="pyarrow")
			assert set(loaded_qa.columns) == {
				"qid",
				"query",
				"retrieval_gt",
				"generation_gt",
			}
			loaded_corpus = pd.read_parquet(corpus_path.name, engine="pyarrow")
			assert set(loaded_corpus.columns) == {"doc_id", "contents", "metadata"}
			corpus_path.close()
			os.unlink(corpus_path.name)
		qa_path.close()
		os.unlink(qa_path.name)



from autorag import embedding_models
from autorag.evaluation.util import cast_metrics


def test_cast_metrics():
	metric1 = ["bleu", "meteor", "rouge"]
	metric_names, metric_params = cast_metrics(metric1)
	assert metric_names == ["bleu", "meteor", "rouge"]
	assert metric_params == [{}, {}, {}]

	metric2 = [
		{"metric_name": "bleu"},
		{"metric_name": "meteor"},
		{"metric_name": "rouge"},
	]
	metric_names, metric_params = cast_metrics(metric2)
	assert metric_names == ["bleu", "meteor", "rouge"]
	assert metric_params == [{}, {}, {}]

	metric3 = [
		{"metric_name": "bleu"},
		{"metric_name": "sem_score", "embedding_model": "openai"},
	]
	metric_names, metric_params = cast_metrics(metric3)
	assert metric_names == ["bleu", "sem_score"]
	assert metric_params == [{}, {"embedding_model": embedding_models["openai"]}]

	metric4 = [
		{"metric_name": "bleu", "extra_param": "extra"},
		{
			"metric_name": "sem_score",
			"embedding_model": "openai",
			"extra_param": "extra",
		},
	]
	metric_names, metric_params = cast_metrics(metric4)
	assert metric_names == ["bleu", "sem_score"]
	assert metric_params == [
		{"extra_param": "extra"},
		{"embedding_model": embedding_models["openai"], "extra_param": "extra"},
	]



from llama_index.core.llms import MockLLM

from autorag.data.qa.evolve.llama_index_query_evolve import (
	conditional_evolve_ragas,
	reasoning_evolve_ragas,
	compress_ragas,
)
from autorag.data.qa.schema import QA
from tests.autorag.data.qa.evolve.base_test_query_evolve import qa_df

llm = MockLLM()
qa = QA(qa_df)


def test_conditional_evolve_ragas():
	new_qa = qa.batch_apply(conditional_evolve_ragas, llm=llm)
	assert "query" in new_qa.data.columns
	assert all(isinstance(query, str) for query in new_qa.data["query"].tolist())
	assert len(new_qa.data) == len(qa_df)
	assert all(
		x != y for x, y in zip(new_qa.data["query"].tolist(), qa_df["query"].tolist())
	)


def test_reasoning_evolve_ragas():
	new_qa = qa.batch_apply(reasoning_evolve_ragas, llm=llm)
	assert "query" in new_qa.data.columns
	assert all(isinstance(query, str) for query in new_qa.data["query"].tolist())
	assert len(new_qa.data) == len(qa_df)
	assert all(
		x != y for x, y in zip(new_qa.data["query"].tolist(), qa_df["query"].tolist())
	)


def test_compress_ragas():
	new_qa = qa.batch_apply(compress_ragas, llm=llm)
	assert "query" in new_qa.data.columns
	assert all(isinstance(query, str) for query in new_qa.data["query"].tolist())
	assert len(new_qa.data) == len(qa_df)
	assert all(
		x != y for x, y in zip(new_qa.data["query"].tolist(), qa_df["query"].tolist())
	)



import logging
import os
from typing import List, Tuple, Optional

import chromadb
import pandas as pd
import torch
from chromadb import GetResult, QueryResult
from chromadb.utils.batch_utils import create_batches
from llama_index.core.embeddings import BaseEmbedding
from llama_index.embeddings.openai import OpenAIEmbedding

from autorag import embedding_models
from autorag.nodes.retrieval.base import evenly_distribute_passages, BaseRetrieval
from autorag.utils import validate_corpus_dataset, cast_corpus_dataset
from autorag.utils.util import (
	get_event_loop,
	process_batch,
	openai_truncate_by_token,
	flatten_apply,
	result_to_dataframe,
	pop_params,
	fetch_contents,
)

logger = logging.getLogger("AutoRAG")


class VectorDB(BaseRetrieval):
	def __init__(self, project_dir: str, embedding_model: str, **kwargs):
		"""
		Initialize VectorDB retrieval node.

		:param project_dir: The project directory path.
		:param embedding_model: The embedding model name.
			It will initialize from the autorag.embedding_models dictionary.
			You can add your own models to the dictionary.
			For more information, see https://docs.auto-rag.com/local_model.html#configure-the-embedding-model
		:param kwargs: The optional arguments.
			Not affected in the init method.
		"""
		super().__init__(project_dir)

		# init chroma collection
		chroma_path = os.path.join(self.resources_dir, "chroma")
		assert (
			chroma_path is not None
		), "chroma_path must be specified for using vectordb retrieval."
		assert os.path.exists(
			chroma_path
		), f"chroma_path {chroma_path} does not exist. Please ingest first."

		self.chroma_collection = load_chroma_collection(
			db_path=chroma_path, collection_name=embedding_model
		)

		# init embedding model
		if embedding_model in embedding_models:
			self.embedding_model = embedding_models[embedding_model]
		else:
			logger.error(f"embedding_model_str {embedding_model} does not exist.")
			raise KeyError(f"embedding_model_str {embedding_model} does not exist.")

	def __del__(self):
		del self.chroma_collection
		del self.embedding_model
		if torch.cuda.is_available():
			torch.cuda.empty_cache()
		super().__del__()

	@result_to_dataframe(["retrieved_contents", "retrieved_ids", "retrieve_scores"])
	def pure(self, previous_result: pd.DataFrame, *args, **kwargs):
		queries = self.cast_to_run(previous_result)
		pure_params = pop_params(self._pure, kwargs)
		ids, scores = self._pure(queries, **pure_params)
		contents = fetch_contents(self.corpus_df, ids)
		return contents, ids, scores

	def _pure(
		self,
		queries: List[List[str]],
		top_k: int,
		embedding_batch: int = 128,
		ids: Optional[List[List[str]]] = None,
	) -> Tuple[List[List[str]], List[List[float]]]:
		"""
		VectorDB retrieval function.
		You have to get a chroma collection that is already ingested.
		You have to get an embedding model that is already used in ingesting.

		:param queries: 2-d list of query strings.
		    Each element of the list is a query strings of each row.
		:param top_k: The number of passages to be retrieved.
		:param embedding_batch: The number of queries to be processed in parallel.
		    This is used to prevent API error at the query embedding.
		    Default is 128.
		:param ids: The optional list of ids that you want to retrieve.
		    You don't need to specify this in the general use cases.
		    Default is None.

		:return: The 2-d list contains a list of passage ids that retrieved from vectordb and 2-d list of its scores.
		    It will be a length of queries. And each element has a length of top_k.
		"""
		# check if bm25_corpus is valid
		assert (
			self.chroma_collection.count() > 0
		), "collection must contain at least one document. Please check you ingested collection correctly."
		# truncate queries and embedding execution here.
		openai_embedding_limit = 8191
		if isinstance(self.embedding_model, OpenAIEmbedding):
			queries = list(
				map(
					lambda query_list: openai_truncate_by_token(
						query_list,
						openai_embedding_limit,
						self.embedding_model.model_name,
					),
					queries,
				)
			)

		query_embeddings = flatten_apply(
			run_query_embedding_batch,
			queries,
			embedding_model=self.embedding_model,
			batch_size=embedding_batch,
		)

		# if ids are specified, fetch the ids score from Chroma
		if ids is not None:
			client = chromadb.Client()
			score_result = list(
				map(
					lambda query_embedding_list, id_list: get_id_scores(
						id_list, query_embedding_list, self.chroma_collection, client
					),
					query_embeddings,
					ids,
				)
			)
			return ids, score_result

		# run async vector_db_pure function
		tasks = [
			vectordb_pure(query_embedding, top_k, self.chroma_collection)
			for query_embedding in query_embeddings
		]
		loop = get_event_loop()
		results = loop.run_until_complete(
			process_batch(tasks, batch_size=embedding_batch)
		)
		id_result = list(map(lambda x: x[0], results))
		score_result = list(map(lambda x: x[1], results))
		return id_result, score_result


async def vectordb_pure(
	query_embeddings: List[List[float]], top_k: int, collection: chromadb.Collection
) -> Tuple[List[str], List[float]]:
	"""
	Async VectorDB retrieval function.
	Its usage is for async retrieval of vector_db row by row.

	:param query_embeddings: A list of query embeddings.
	:param top_k: The number of passages to be retrieved.
	:param collection: A chroma collection instance that will be used to retrieve passages.
	:return: The tuple contains a list of passage ids that are retrieved from vectordb and a list of its scores.
	"""
	id_result, score_result = [], []
	for embedded_query in query_embeddings:
		result = collection.query(query_embeddings=embedded_query, n_results=top_k)
		id_result.extend(result["ids"])
		score_result.extend(
			list(map(lambda lst: list(map(lambda x: 1 - x, lst)), result["distances"]))
		)

	# Distribute passages evenly
	id_result, score_result = evenly_distribute_passages(id_result, score_result, top_k)
	# sort id_result and score_result by score
	result = [
		(_id, score)
		for score, _id in sorted(
			zip(score_result, id_result), key=lambda pair: pair[0], reverse=True
		)
	]
	id_result, score_result = zip(*result)
	return list(id_result), list(score_result)


def vectordb_ingest(
	collection: chromadb.Collection,
	corpus_data: pd.DataFrame,
	embedding_model: BaseEmbedding,
	embedding_batch: int = 128,
):
	"""
	Ingest given corpus data to the chromadb collection.
	It truncates corpus content when the embedding model is OpenAIEmbedding to the 8191 tokens.
	Plus, when the corpus content is empty (whitespace), it will be ignored.
	And if there is a document id that already exists in the collection, it will be ignored.

	:param collection: Chromadb collection instance to ingest.
	:param corpus_data: The corpus data that contains doc_id and contents columns.
	:param embedding_model: An embedding model instance that will be used to embed queries.
	:param embedding_batch: The number of chunks that will be processed in parallel.
	"""
	embedding_model.embed_batch_size = embedding_batch
	corpus_data = cast_corpus_dataset(corpus_data)
	validate_corpus_dataset(corpus_data)
	ids = corpus_data["doc_id"].tolist()

	# Query the collection to check if IDs already exist
	existing_ids = set(
		collection.get(ids=ids)["ids"]
	)  # Assuming 'ids' is the key in the response
	new_passage = corpus_data[~corpus_data["doc_id"].isin(existing_ids)]

	if not new_passage.empty:
		new_contents = new_passage["contents"].tolist()

		# truncate by token if embedding_model is OpenAIEmbedding
		if isinstance(embedding_model, OpenAIEmbedding):
			openai_embedding_limit = (
				8191  # all openai embedding models have 8191 max token input
			)
			new_contents = openai_truncate_by_token(
				new_contents, openai_embedding_limit, embedding_model.model_name
			)

		new_ids = new_passage["doc_id"].tolist()
		embedded_contents = embedding_model.get_text_embedding_batch(
			new_contents, show_progress=True
		)
		input_batches = create_batches(
			api=collection._client, ids=new_ids, embeddings=embedded_contents
		)
		for batch in input_batches:
			ids = batch[0]
			embed_content = batch[1]
			collection.add(ids=ids, embeddings=embed_content)


def run_query_embedding_batch(
	queries: List[str], embedding_model: BaseEmbedding, batch_size: int
) -> List[List[float]]:
	result = []
	for i in range(0, len(queries), batch_size):
		batch = queries[i : i + batch_size]
		embeddings = embedding_model.get_text_embedding_batch(batch)
		result.extend(embeddings)
	return result


def get_id_scores(
	ids: List[str],
	query_embeddings: List[List[float]],
	collection: chromadb.Collection,
	temp_client: chromadb.Client,
) -> List[float]:
	if len(ids) == 0 or ids is None or not bool(ids):
		return []

	id_results: GetResult = collection.get(ids, include=["embeddings"])
	temp_collection = temp_client.create_collection(
		name="temp", metadata={"hnsw:space": "cosine"}
	)
	temp_collection.add(ids=id_results["ids"], embeddings=id_results["embeddings"])

	query_result: QueryResult = temp_collection.query(
		query_embeddings=query_embeddings, n_results=len(ids)
	)
	assert len(query_result["ids"]) == len(query_result["distances"])
	id_scores_dict = {id_: [] for id_ in ids}
	for id_list, score_list in zip(query_result["ids"], query_result["distances"]):
		for id_ in list(id_scores_dict.keys()):
			id_idx = id_list.index(id_)
			id_scores_dict[id_].append(score_list[id_idx])
	id_scores_pd = pd.DataFrame(id_scores_dict)
	temp_client.delete_collection("temp")
	return id_scores_pd.max(axis=0).tolist()


def load_chroma_collection(db_path: str, collection_name: str) -> chromadb.Collection:
	db = chromadb.PersistentClient(path=db_path)
	collection = db.get_collection(name=collection_name)
	return collection



from typing import List, Tuple, Iterable

import pandas as pd
import torch
from FlagEmbedding import FlagReranker
from tqdm import tqdm

from autorag.nodes.passagereranker.base import BasePassageReranker
from autorag.utils.util import (
	make_batch,
	sort_by_scores,
	flatten_apply,
	select_top_k,
	pop_params,
	result_to_dataframe,
)


class FlagEmbeddingReranker(BasePassageReranker):
	def __init__(
		self, project_dir, model_name: str = "BAAI/bge-reranker-large", *args, **kwargs
	):
		"""
		Initialize the FlagEmbeddingReranker module.

		:param project_dir: The project directory.
		:param model_name: The name of the BAAI Reranker normal-model name.
		Default is "BAAI/bge-reranker-large"
		:param kwargs: Extra parameter for FlagEmbedding.FlagReranker
		"""
		super().__init__(project_dir)
		model_params = pop_params(FlagReranker.__init__, kwargs)
		model_params.pop("model_name_or_path", None)
		self.model = FlagReranker(model_name_or_path=model_name, **model_params)

	def __del__(self):
		del self.model
		if torch.cuda.is_available():
			torch.cuda.empty_cache()
		super().__del__()

	@result_to_dataframe(["retrieved_contents", "retrieved_ids", "retrieve_scores"])
	def pure(self, previous_result: pd.DataFrame, *args, **kwargs):
		queries, contents, _, ids = self.cast_to_run(previous_result)
		top_k = kwargs.pop("top_k")
		batch = kwargs.pop("batch", 64)
		return self._pure(queries, contents, ids, top_k, batch)

	def _pure(
		self,
		queries: List[str],
		contents_list: List[List[str]],
		ids_list: List[List[str]],
		top_k: int,
		batch: int = 64,
	) -> Tuple[List[List[str]], List[List[str]], List[List[float]]]:
		"""
		Rerank a list of contents based on their relevance to a query using BAAI normal-Reranker model.

		:param queries: The list of queries to use for reranking
		:param contents_list: The list of lists of contents to rerank
		:param ids_list: The list of lists of ids retrieved from the initial ranking
		:param top_k: The number of passages to be retrieved
		:param batch: The number of queries to be processed in a batch
			Default is 64.
		:return: Tuple of lists containing the reranked contents, ids, and scores
		"""
		nested_list = [
			list(map(lambda x: [query, x], content_list))
			for query, content_list in zip(queries, contents_list)
		]
		rerank_scores = flatten_apply(
			flag_embedding_run_model, nested_list, model=self.model, batch_size=batch
		)

		df = pd.DataFrame(
			{
				"contents": contents_list,
				"ids": ids_list,
				"scores": rerank_scores,
			}
		)
		df[["contents", "ids", "scores"]] = df.apply(
			sort_by_scores, axis=1, result_type="expand"
		)
		results = select_top_k(df, ["contents", "ids", "scores"], top_k)

		return (
			results["contents"].tolist(),
			results["ids"].tolist(),
			results["scores"].tolist(),
		)


def flag_embedding_run_model(input_texts, model, batch_size: int):
	batch_input_texts = make_batch(input_texts, batch_size)
	results = []
	for batch_texts in tqdm(batch_input_texts):
		with torch.no_grad():
			pred_scores = model.compute_score(sentence_pairs=batch_texts)
		if batch_size == 1 or not isinstance(pred_scores, Iterable):
			results.append(pred_scores)
		else:
			results.extend(pred_scores)
	return results



from unittest.mock import patch

import openai.resources.chat
import pandas as pd
import pytest

from autorag.nodes.generator import OpenAILLM
from tests.autorag.nodes.generator.test_generator_base import (
	prompts,
	check_generated_texts,
	check_generated_tokens,
	check_generated_log_probs,
)
from tests.mock import mock_openai_chat_create


@pytest.fixture
def openai_llm_instance():
	return OpenAILLM(
		project_dir=".", llm="gpt-3.5-turbo", api_key="mock_openai_api_key"
	)


@patch.object(
	openai.resources.chat.completions.AsyncCompletions,
	"create",
	mock_openai_chat_create,
)
def test_openai_llm(openai_llm_instance):
	answers, tokens, log_probs = openai_llm_instance._pure(
		prompts, temperature=0.5, logprobs=False, n=3
	)
	check_generated_texts(answers)
	check_generated_tokens(tokens)
	check_generated_log_probs(log_probs)


@patch.object(
	openai.resources.chat.completions.AsyncCompletions,
	"create",
	mock_openai_chat_create,
)
def test_openai_llm_node():
	previous_result = pd.DataFrame(
		{"prompts": prompts, "qid": ["id-1", "id-2", "id-3"]}
	)
	result_df = OpenAILLM.run_evaluator(
		project_dir=".",
		previous_result=previous_result,
		llm="gpt-4o-mini",
		api_key="mock_openai_api_key",
		temperature=0.5,
	)
	check_generated_texts(result_df["generated_texts"].tolist())
	check_generated_tokens(result_df["generated_tokens"].tolist())
	check_generated_log_probs(result_df["generated_log_probs"].tolist())


@patch.object(
	openai.resources.chat.completions.AsyncCompletions,
	"create",
	mock_openai_chat_create,
)
def test_openai_llm_truncate(openai_llm_instance):
	prompt = [
		f"havertz on the block and I am {i}th player on the Arsenal."
		for i in range(50_000)
	]
	prompt = " ".join(prompt)
	answers, tokens, log_probs = openai_llm_instance._pure([prompt] * 3)
	check_generated_texts(answers)
	check_generated_tokens(tokens)
	check_generated_log_probs(log_probs)



from unittest.mock import patch

import pytest
from llama_index.embeddings.openai import OpenAIEmbedding

from autorag.nodes.passagefilter import SimilarityPercentileCutoff
from tests.autorag.nodes.passagefilter.test_passage_filter_base import (
	queries_example,
	contents_example,
	scores_example,
	ids_example,
	base_passage_filter_test,
	project_dir,
	previous_result,
	base_passage_filter_node_test,
)
from tests.mock import mock_get_text_embedding_batch


@pytest.fixture
def similarity_percentile_cutoff_instance():
	return SimilarityPercentileCutoff(
		project_dir=project_dir,
		previous_result=previous_result,
		embedding_model="openai_embed_3_large",
	)


@patch.object(
	OpenAIEmbedding,
	"get_text_embedding_batch",
	mock_get_text_embedding_batch,
)
def test_similarity_percentile_cutoff(similarity_percentile_cutoff_instance):
	assert (
		similarity_percentile_cutoff_instance.embedding_model.model_name
		== "text-embedding-3-large"
	)
	contents, ids, scores = similarity_percentile_cutoff_instance._pure(
		queries_example,
		contents_example,
		scores_example,
		ids_example,
		percentile=0.85,
		batch=64,
	)
	num_top_k = int(len(contents_example[0]) * 0.85)
	assert len(contents[0]) == len(contents[1]) == num_top_k
	base_passage_filter_test(contents, ids, scores)


@patch.object(
	OpenAIEmbedding,
	"get_text_embedding_batch",
	mock_get_text_embedding_batch,
)
def test_similarity_percentile_cutoff_node():
	result_df = SimilarityPercentileCutoff.run_evaluator(
		project_dir=project_dir, previous_result=previous_result, percentile=0.9
	)
	base_passage_filter_node_test(result_df)



import os
import pathlib
import shutil
import tempfile
import uuid
from datetime import datetime
from unittest.mock import patch

import chromadb
import pandas as pd
import pytest
from llama_index.core import MockEmbedding
from llama_index.embeddings.openai import OpenAIEmbedding

from autorag.nodes.retrieval import VectorDB
from autorag.nodes.retrieval.vectordb import vectordb_ingest, get_id_scores
from tests.autorag.nodes.retrieval.test_retrieval_base import (
	queries,
	corpus_df,
	previous_result,
	base_retrieval_test,
	base_retrieval_node_test,
	searchable_input_ids,
)
from tests.mock import mock_get_text_embedding_batch

root_dir = pathlib.PurePath(
	os.path.dirname(os.path.realpath(__file__))
).parent.parent.parent
resource_path = os.path.join(root_dir, "resources")

embedding_model = MockEmbedding(1536)


@pytest.fixture
def ingested_vectordb():
	with tempfile.TemporaryDirectory(ignore_cleanup_errors=True) as chroma_path:
		db = chromadb.PersistentClient(path=chroma_path)
		collection = db.create_collection(
			name="test_vectordb_retrieval", metadata={"hnsw:space": "cosine"}
		)

		vectordb_ingest(collection, corpus_df, embedding_model)

		assert collection.count() == 5
		yield collection


@pytest.fixture
def empty_chromadb():
	with tempfile.TemporaryDirectory(ignore_cleanup_errors=True) as chroma_path:
		db = chromadb.PersistentClient(path=chroma_path)
		collection = db.create_collection(
			name="test_vectordb_retrieval", metadata={"hnsw:space": "cosine"}
		)

		yield collection


@pytest.fixture
def project_dir_for_vectordb_node():
	with tempfile.TemporaryDirectory(ignore_cleanup_errors=True) as test_project_dir:
		os.makedirs(os.path.join(test_project_dir, "resources"))
		chroma_path = os.path.join(test_project_dir, "resources", "chroma")
		os.makedirs(chroma_path)
		db = chromadb.PersistentClient(path=chroma_path)
		collection = db.create_collection(
			name="openai", metadata={"hnsw:space": "cosine"}
		)
		os.makedirs(os.path.join(test_project_dir, "data"))
		corpus_path = os.path.join(test_project_dir, "data", "corpus.parquet")
		corpus_df.to_parquet(corpus_path, index=False)
		vectordb_ingest(collection, corpus_df, embedding_model)

		yield test_project_dir


@pytest.fixture
def project_dir_for_vectordb_node_from_sample_project():
	with tempfile.TemporaryDirectory(ignore_cleanup_errors=True) as test_project_dir:
		sample_project_dir = os.path.join(resource_path, "sample_project")
		# copy & paste all folders and files in the sample_project folder
		shutil.copytree(sample_project_dir, test_project_dir, dirs_exist_ok=True)

		chroma_path = os.path.join(test_project_dir, "resources", "chroma")
		os.makedirs(chroma_path)
		db = chromadb.PersistentClient(path=chroma_path)
		collection = db.create_collection(
			name="openai", metadata={"hnsw:space": "cosine"}
		)
		corpus_path = os.path.join(test_project_dir, "data", "corpus.parquet")
		local_corpus_df = pd.read_parquet(corpus_path, engine="pyarrow")
		vectordb_ingest(collection, local_corpus_df, embedding_model)

		yield test_project_dir


@pytest.fixture
def vectordb_instance(project_dir_for_vectordb_node):
	vectordb = VectorDB(
		project_dir=project_dir_for_vectordb_node, embedding_model="openai"
	)
	yield vectordb


@patch.object(
	OpenAIEmbedding,
	"get_text_embedding_batch",
	mock_get_text_embedding_batch,
)
def test_vectordb_retrieval(vectordb_instance):
	top_k = 4
	id_result, score_result = vectordb_instance._pure(
		queries,
		top_k=top_k,
	)
	base_retrieval_test(id_result, score_result, top_k)


@patch.object(
	OpenAIEmbedding,
	"get_text_embedding_batch",
	mock_get_text_embedding_batch,
)
def test_vectordb_retrieval_ids(vectordb_instance):
	ids = [["doc2", "doc3"], ["doc1", "doc2"], ["doc4", "doc5"]]
	id_result, score_result = vectordb_instance._pure(
		queries,
		top_k=4,
		ids=ids,
	)
	assert id_result == ids
	assert len(id_result) == len(score_result) == 3
	assert all([len(score_list) == 2 for score_list in score_result])


@patch.object(
	OpenAIEmbedding,
	"get_text_embedding_batch",
	mock_get_text_embedding_batch,
)
def test_vectordb_retrieval_ids_empty(vectordb_instance):
	ids = [["doc2", "doc3"], [], ["doc4"]]
	id_result, score_result = vectordb_instance._pure(
		queries,
		top_k=4,
		ids=ids,
	)
	assert id_result == ids
	assert len(id_result) == len(score_result) == 3
	assert len(score_result[0]) == 2
	assert len(score_result[1]) == 0
	assert len(score_result[2]) == 1


@patch.object(
	OpenAIEmbedding,
	"get_text_embedding_batch",
	mock_get_text_embedding_batch,
)
def test_vectordb_node(project_dir_for_vectordb_node_from_sample_project):
	result_df = VectorDB.run_evaluator(
		project_dir=project_dir_for_vectordb_node_from_sample_project,
		previous_result=previous_result,
		top_k=4,
		embedding_model="openai",
	)
	base_retrieval_node_test(result_df)


@patch.object(
	OpenAIEmbedding,
	"get_text_embedding_batch",
	mock_get_text_embedding_batch,
)
def test_vectordb_node_ids(project_dir_for_vectordb_node_from_sample_project):
	result_df = VectorDB.run_evaluator(
		project_dir=project_dir_for_vectordb_node_from_sample_project,
		previous_result=previous_result,
		top_k=4,
		embedding_model="openai",
		ids=searchable_input_ids,
	)
	contents = result_df["retrieved_contents"].tolist()
	ids = result_df["retrieved_ids"].tolist()
	scores = result_df["retrieve_scores"].tolist()
	assert len(contents) == len(ids) == len(scores) == 5
	assert len(contents[0]) == len(ids[0]) == len(scores[0]) == 2
	assert ids[0] == searchable_input_ids[0]


@patch.object(
	OpenAIEmbedding,
	"get_text_embedding_batch",
	mock_get_text_embedding_batch,
)
def test_duplicate_id_vectordb_ingest(ingested_vectordb):
	vectordb_ingest(ingested_vectordb, corpus_df, embedding_model)
	assert ingested_vectordb.count() == 5

	new_doc_id = ["doc4", "doc5", "doc6", "doc7", "doc8"]
	new_contents = [
		"This is a test document 4.",
		"This is a test document 5.",
		"This is a test document 6.",
		"This is a test document 7.",
		"This is a test document 8.",
	]
	new_metadata = [{"datetime": datetime.now()} for _ in range(5)]
	new_corpus_df = pd.DataFrame(
		{"doc_id": new_doc_id, "contents": new_contents, "metadata": new_metadata}
	)
	vectordb_ingest(ingested_vectordb, new_corpus_df, embedding_model)

	assert ingested_vectordb.count() == 8


@patch.object(
	OpenAIEmbedding,
	"get_text_embedding_batch",
	mock_get_text_embedding_batch,
)
def test_long_text_vectordb_ingest(ingested_vectordb):
	new_doc_id = ["doc6", "doc7"]
	new_contents = ["This is a test" * 20000, "This is a test" * 40000]
	new_metadata = [{"datetime": datetime.now()} for _ in range(2)]
	new_corpus_df = pd.DataFrame(
		{"doc_id": new_doc_id, "contents": new_contents, "metadata": new_metadata}
	)
	assert isinstance(embedding_model, MockEmbedding)
	vectordb_ingest(ingested_vectordb, new_corpus_df, embedding_model)

	assert ingested_vectordb.count() == 7


def mock_get_text_embedding_batch(self, texts, **kwargs):
	return [[3.0, 4.1, 3.2] for _ in range(len(texts))]


@patch.object(
	OpenAIEmbedding, "get_text_embedding_batch", mock_get_text_embedding_batch
)
def test_long_ids_ingest(empty_chromadb):
	embedding_model = OpenAIEmbedding()
	content_df = pd.DataFrame(
		{
			"doc_id": [str(uuid.uuid4()) for _ in range(10000)],
			"contents": ["havertz" for _ in range(10000)],
			"metadata": [
				{"last_modified_datetime": datetime.now()} for _ in range(10000)
			],
		}
	)
	vectordb_ingest(empty_chromadb, content_df, embedding_model)


def test_get_id_scores(ingested_vectordb):
	ids = ["doc2", "doc3", "doc4"]
	embedding_model = MockEmbedding(1536)
	queries = [
		"다이노스 오! 권희동~ 엔씨 오 권희동 오 권희동 권희동 안타~",
		"두산의 헨리 라모스 오오오 라모스 시원하게 화끈하게 날려버려라",
	]
	query_embeddings = embedding_model.get_text_embedding_batch(queries)
	client = chromadb.Client()
	scores = get_id_scores(ids, query_embeddings, ingested_vectordb, client)
	assert len(scores) == 3
	assert isinstance(scores[0], float)



import pytest
from sklearn.utils.fixes import percentile

from autorag.nodes.passagefilter import percentile_cutoff, PercentileCutoff
from tests.autorag.nodes.passagefilter.test_passage_filter_base import (
	queries_example,
	contents_example,
	scores_example,
	ids_example,
	base_passage_filter_test,
	project_dir,
	previous_result,
	base_passage_filter_node_test,
)


@pytest.fixture
def percentile_cutoff_instance():
	return PercentileCutoff(
		project_dir=project_dir, previous_result=previous_result, percentile=0.6
	)


def test_percentile_cutoff(percentile_cutoff_instance):
	contents, ids, scores = percentile_cutoff_instance._pure(
		queries_example, contents_example, scores_example, ids_example, percentile=0.6
	)
	base_passage_filter_test(contents, ids, scores)
	assert scores[0] == [0.8, 0.5]
	assert contents[0] == [
		"Paris is the capital of France.",
		"Paris is one of the capital from France. Isn't it?",
	]


def test_percentile_cutoff_reverse(percentile_cutoff_instance):
	contents, ids, scores = percentile_cutoff_instance._pure(
		queries_example,
		contents_example,
		scores_example,
		ids_example,
		percentile=0.6,
		reverse=True,
	)
	base_passage_filter_test(contents, ids, scores)
	assert scores[0] == [0.1, 0.1]
	assert contents[0] == ["NomaDamas is Great Team", "havertz is suck at soccer"]


def test_percentile_cutoff_node():
	result_df = PercentileCutoff.run_evaluator(
		project_dir=project_dir, previous_result=previous_result, percentile=0.9
	)
	base_passage_filter_node_test(result_df)



from typing import Tuple, List

import pandas as pd
import pytest

from autorag.evaluation import evaluate_retrieval_contents
from autorag.schema.metricinput import MetricInput

gt = [
	[["Enough for drinking water", "Just looking for a water bottle"]],
	[["Do you want to buy some?"]],
]
pred = [
	[
		"Enough for mixing water",
		"I want to do a nothing",
		"Just looking is a very healthy",
	],
	["Do you want to buy some?", "I want to buy some", "I want to buy some water"],
]


@evaluate_retrieval_contents(
	metric_inputs=[MetricInput(retrieval_gt_contents=gt_contents) for gt_contents in gt],
	metrics=[
		"retrieval_token_recall",
		"retrieval_token_precision",
		"retrieval_token_f1",
	],
)
def pseudo_retrieval() -> Tuple[List[List[str]], List[List[float]], List[List[str]]]:
	return (
		pred,
		[[0.3, 0.2, 0.1], [0.3, 0.2, 0.1]],
		[["pred-0", "pred-1", "pred-2"], ["pred-3", "pred-4", "pred-5"]],
	)


def test_evaluate_retrieval_contents():
	result_df = pseudo_retrieval()
	assert isinstance(result_df, pd.DataFrame)
	assert len(result_df) == 2
	assert set(result_df.columns) == {
		"retrieved_contents",
		"retrieved_ids",
		"retrieve_scores",
		"retrieval_token_recall",
		"retrieval_token_precision",
		"retrieval_token_f1",
	}
	assert result_df["retrieval_token_recall"].tolist() == pytest.approx(
		[0.383333, 0.777777], rel=0.001
	)
	assert result_df["retrieval_token_precision"].tolist() == pytest.approx(
		[0.383333, 0.8222222], rel=0.001
	)
	assert result_df["retrieval_token_f1"].tolist() == pytest.approx(
		[0.38333, 0.797979], rel=0.001
	)



import os
from unittest.mock import patch

import openai
import pandas as pd
import pytest
from llama_index.embeddings.openai import OpenAIEmbedding
from openai.types.chat import (
	ChatCompletion,
	ChatCompletionMessage,
	ChatCompletionTokenLogprob,
)
from openai.types.chat.chat_completion import Choice, ChoiceLogprobs
from openai.types.chat.chat_completion_token_logprob import TopLogprob
from transformers import AutoTokenizer

from autorag.evaluation.generation import evaluate_generation
from autorag.schema.metricinput import MetricInput
from tests.mock import mock_get_text_embedding_batch

generation_gts = [
	["The dog had bit the man.", "The man had bitten the dog."],
	["I want to be a artist, but I end up to be a programmer."],
	[
		"To be a artist these days, you can overcome by AI.",
		"To be a programmer these days, you can overcome by AI.",
		"To be a lawyer these days, you can overcome by AI.",
	],
]
pseudo_generations = [
	"The dog bit the man.",
	"It really like to be a programmer, but I think artist is my passion.",
	"To be a artist these days, you can overcome by AI.",
]

tokenizer = AutoTokenizer.from_pretrained("gpt2")
pseudo_tokens = list(map(lambda x: tokenizer.tokenize(x), pseudo_generations))
pseudo_log_probs = list(map(lambda x: [0.1] * len(x), pseudo_tokens))


@evaluate_generation(
	metric_inputs=[MetricInput(generation_gt=gen_gt) for gen_gt in generation_gts],
	metrics=["bleu", "meteor", "rouge", "sem_score", "g_eval"],
)
def pseudo_generation():
	return pseudo_generations


@evaluate_generation(
	metric_inputs=[MetricInput(generation_gt=gen_gt) for gen_gt in generation_gts],
	metrics=["bleu", "meteor", "donggeon_metric"]
)
def pseudo_generation_with_log_probs():
	return pseudo_generations, pseudo_tokens, pseudo_log_probs


@evaluate_generation(
	metric_inputs=[MetricInput(generation_gt=gen_gt) for gen_gt in generation_gts],
	metrics=[
		{"metric_name": "bleu"},
		{"metric_name": "sem_score", "embedding_model": "openai"},
	],
)
def pseudo_generation_dict_metrics():
	return pseudo_generations


async def mock_g_eval_openai_create(*args, **kwargs):
	sample_choice = Choice(
		finish_reason="stop",
		index=0,
		message=ChatCompletionMessage(
			content="2",
			role="assistant",
		),
		logprobs=ChoiceLogprobs(
			content=[
				ChatCompletionTokenLogprob(
					token="2",
					logprob=2.8,
					top_logprobs=[
						TopLogprob(token="2", logprob=2.8),
					],
				)
			],
		),
	)
	if "n" not in kwargs.keys():
		n = 1
	else:
		n = kwargs["n"]
	return ChatCompletion(
		id="_id",
		choices=[sample_choice] * n,
		created=1713363661,
		model=kwargs["model"],
		object="chat.completion",
	)


@patch.object(
	openai.resources.chat.completions.AsyncCompletions,
	"create",
	mock_g_eval_openai_create,
)
@patch.object(
	OpenAIEmbedding,
	"get_text_embedding_batch",
	mock_get_text_embedding_batch,
)
def test_evaluate_generation():
	os.environ["OPENAI_API_KEY"] = "mock_openai_api_key"
	result_df = pseudo_generation()
	assert isinstance(result_df, pd.DataFrame)
	assert len(result_df) == 3
	assert len(result_df.columns) == 6
	assert set(result_df.columns) == {
		"generated_texts",
		"bleu",
		"meteor",
		"rouge",
		"sem_score",
		"g_eval",
	}

	with pytest.warns():
		result_df_log_probs = pseudo_generation_with_log_probs()
	assert isinstance(result_df_log_probs, pd.DataFrame)
	assert len(result_df_log_probs) == 3
	assert len(result_df_log_probs.columns) == 5
	assert set(result_df_log_probs.columns) == {
		"generated_texts",
		"bleu",
		"meteor",
		"generated_tokens",
		"generated_log_probs",
	}

	assert result_df_log_probs["generated_texts"].tolist() == pseudo_generations
	assert result_df_log_probs["generated_tokens"].tolist() == pseudo_tokens
	assert result_df_log_probs["generated_log_probs"].tolist() == pseudo_log_probs

	assert all(
		list(
			map(
				lambda x: x[0] == pytest.approx(x[1], 0.001),
				zip(result_df["bleu"].tolist(), [51.1507, 23.5783, 100.0]),
			)
		)
	)
	assert all(
		list(
			map(
				lambda x: x[0] == pytest.approx(x[1], 0.001),
				zip(result_df["meteor"].tolist(), [0.853462, 0.5859375, 1.0]),
			)
		)
	)
	assert all(
		list(
			map(
				lambda x: x[0] == pytest.approx(x[1], 0.001),
				zip(result_df["rouge"].tolist(), [0.909, 0.35714, 1.0]),
			)
		)
	)

	result_df = pseudo_generation_dict_metrics()
	assert isinstance(result_df, pd.DataFrame)
	assert len(result_df) == 3
	assert len(result_df.columns) == 3
	assert set(result_df.columns) == {"generated_texts", "bleu", "sem_score"}



import os
import pathlib
import tempfile

import pandas as pd
import pytest
from guidance import models
from llama_index.core import SimpleDirectoryReader

from autorag.data.legacy.corpus.llama_index import llama_documents_to_parquet
from autorag.data.legacy.qacreation.simple import (
	generate_simple_qa_dataset,
	generate_qa_row,
)
from autorag.data.utils.util import get_file_metadata
from autorag.utils.preprocess import validate_qa_dataset, validate_corpus_dataset
from tests.delete_tests import is_github_action

root_dir = pathlib.PurePath(
	os.path.dirname(os.path.realpath(__file__))
).parent.parent.parent.parent

raw_dir = os.path.join(root_dir, "resources", "data_creation", "raw_dir")

load_file_name = "test_corpus.parquet"


@pytest.fixture
def load_dir():
	with tempfile.TemporaryDirectory(ignore_cleanup_errors=True) as temp_dir:
		yield temp_dir


@pytest.fixture
def output_filedir():
	with tempfile.TemporaryDirectory(ignore_cleanup_errors=True) as temp_dir:
		yield temp_dir


@pytest.mark.skipif(
	is_github_action(),
	reason="Skipping this test on GitHub Actions because it will be deprecated.",
)
def test_generate_simple_qa_dataset(load_dir, output_filedir):
	loader = SimpleDirectoryReader(file_metadata=get_file_metadata, input_dir=raw_dir)

	documents = loader.load_data(show_progress=True, num_workers=0)
	llama_documents_to_parquet(
		llama_documents=documents,
		output_filepath=os.path.join(load_dir, load_file_name),
	)
	assert os.path.exists(os.path.join(load_dir, load_file_name))
	corpus_data = pd.read_parquet(os.path.join(load_dir, load_file_name))
	validate_corpus_dataset(corpus_data)

	qa_dataset = generate_simple_qa_dataset(
		corpus_data=pd.read_parquet(os.path.join(load_dir, load_file_name)),
		llm=models.OpenAI("gpt-3.5-turbo"),
		output_filepath=os.path.join(output_filedir, "qa.parquet"),
		generate_row_function=generate_qa_row,
	)
	validate_qa_dataset(qa_dataset)
	assert os.path.exists(os.path.join(output_filedir, "qa.parquet"))



import pandas as pd
import pytest

from autorag.nodes.generator import Vllm
from tests.autorag.nodes.generator.test_generator_base import (
	prompts,
	check_generated_texts,
	check_generated_tokens,
	check_generated_log_probs,
)


@pytest.mark.skip(reason="This test needs CUDA enabled machine with vllm installed.")
def test_vllm():
	previous_result = pd.DataFrame(
		{"prompts": prompts, "qid": ["id-1", "id-2", "id-3"]}
	)
	result_df = Vllm.run_evaluator(
		project_dir=".",
		previous_result=previous_result,
		tensor_parallel_size=1,
		llm="facebook/opt-125m",
		max_tokens=5,
		temperature=0.5,
	)
	tokens = result_df["generated_tokens"].tolist()
	log_probs = result_df["generated_log_probs"].tolist()
	check_generated_texts(result_df["generated_texts"].tolist())
	check_generated_tokens(tokens)
	check_generated_log_probs(log_probs)
	assert all(len(tokens[i]) == len(log_probs[i]) for i in range(len(tokens)))



import os
import pathlib
import tempfile

import pandas as pd
import pytest
import yaml

from autorag.deploy import (
	summary_df_to_yaml,
	extract_best_config,
	Runner,
	extract_node_line_names,
	extract_node_strategy,
)
from autorag.evaluator import Evaluator
from tests.delete_tests import is_github_action

root_dir = pathlib.PurePath(os.path.dirname(os.path.realpath(__file__))).parent
resource_dir = os.path.join(root_dir, "resources")


@pytest.fixture
def evaluator():
	with tempfile.TemporaryDirectory(ignore_cleanup_errors=True) as project_dir:
		evaluator = Evaluator(
			os.path.join(resource_dir, "qa_data_sample.parquet"),
			os.path.join(resource_dir, "corpus_data_sample.parquet"),
			project_dir=project_dir,
		)
		yield evaluator


@pytest.fixture
def full_config():
	yaml_path = os.path.join(resource_dir, "full.yaml")
	with open(yaml_path, "r") as f:
		yaml_dict = yaml.safe_load(f)
	return yaml_dict


summary_df = pd.DataFrame(
	{
		"node_line_name": ["node_line_2", "node_line_2", "node_line_1"],
		"node_type": ["retrieval", "rerank", "generation"],
		"best_module_filename": [
			"bm25=>top_k_50.parquet",
			"upr=>model_llama-2-havertz_chelsea.parquet",
			"gpt-4=>top_p_0.9.parquet",
		],
		"best_module_name": ["bm25", "upr", "gpt-4"],
		"best_module_params": [
			{"top_k": 50},
			{"model": "llama-2", "havertz": "chelsea"},
			{"top_p": 0.9},
		],
		"best_execution_time": [1.0, 0.5, 2.0],
	}
)
solution_dict = {
	"node_lines": [
		{
			"node_line_name": "node_line_2",
			"nodes": [
				{
					"node_type": "retrieval",
					"strategy": {
						"metrics": [
							"retrieval_f1",
							"retrieval_recall",
							"retrieval_precision",
						],
					},
					"modules": [{"module_type": "bm25", "top_k": 50}],
				},
				{
					"node_type": "rerank",
					"strategy": {
						"metrics": [
							"retrieval_f1",
							"retrieval_recall",
							"retrieval_precision",
						],
						"speed_threshold": 10,
					},
					"modules": [
						{"module_type": "upr", "model": "llama-2", "havertz": "chelsea"}
					],
				},
			],
		},
		{
			"node_line_name": "node_line_1",
			"nodes": [
				{
					"node_type": "generation",
					"strategy": {
						"metrics": ["bleu", "rouge"],
					},
					"modules": [{"module_type": "gpt-4", "top_p": 0.9}],
				}
			],
		},
	]
}


@pytest.fixture
def pseudo_trial_path():
	with tempfile.TemporaryDirectory(ignore_cleanup_errors=True) as project_dir:
		trial_path = os.path.join(project_dir, "0")
		os.makedirs(trial_path)
		summary_df.to_csv(os.path.join(trial_path, "summary.csv"), index=False)
		with open(os.path.join(trial_path, "config.yaml"), "w") as f:
			yaml.safe_dump(solution_dict, f)
		yield trial_path


def test_extract_node_line_names(full_config):
	node_line_names = extract_node_line_names(full_config)
	assert node_line_names == [
		"pre_retrieve_node_line",
		"retrieve_node_line",
		"post_retrieve_node_line",
	]


def test_extract_node_strategy(full_config):
	node_strategies = extract_node_strategy(full_config)
	assert set(list(node_strategies.keys())) == {
		"query_expansion",
		"retrieval",
		"passage_reranker",
		"prompt_maker",
		"generator",
	}
	assert node_strategies["retrieval"] == {
		"metrics": ["retrieval_f1", "retrieval_recall", "retrieval_precision"],
		"speed_threshold": 10,
	}


def test_summary_df_to_yaml():
	yaml_dict = summary_df_to_yaml(summary_df, solution_dict)
	assert yaml_dict == solution_dict


def test_extract_best_config(pseudo_trial_path):
	yaml_dict = extract_best_config(pseudo_trial_path)
	assert yaml_dict == solution_dict
	with tempfile.NamedTemporaryFile(
		suffix="yaml", mode="w+t", delete=False
	) as yaml_path:
		yaml_dict = extract_best_config(pseudo_trial_path, yaml_path.name)
		assert yaml_dict == solution_dict
		assert os.path.exists(yaml_path.name)
		yaml_dict = yaml.safe_load(yaml_path)
		assert yaml_dict == solution_dict
		yaml_path.close()
		os.unlink(yaml_path.name)


def test_runner(evaluator):
	evaluator.start_trial(os.path.join(resource_dir, "simple_mock.yaml"))
	project_dir = evaluator.project_dir

	def runner_test(runner: Runner):
		answer = runner.run(
			"What is the best movie in Korea? Have Korea movie ever won Oscar?",
			"retrieved_contents",
		)
		assert len(answer) == 10
		assert isinstance(answer, list)
		assert isinstance(answer[0], str)

	runner = Runner.from_trial_folder(os.path.join(project_dir, "0"))
	runner_test(runner)
	runner_test(runner)

	with tempfile.NamedTemporaryFile(
		suffix="yaml", mode="w+t", delete=False
	) as yaml_path:
		extract_best_config(os.path.join(project_dir, "0"), yaml_path.name)
		runner = Runner.from_yaml(yaml_path.name, project_dir=project_dir)
		runner_test(runner)
		yaml_path.close()
		os.unlink(yaml_path.name)


@pytest.mark.skipif(is_github_action(), reason="Skipping this test on GitHub Actions")
def test_runner_full(evaluator):
	runner = Runner.from_trial_folder(os.path.join(resource_dir, "result_project", "0"))
	answer = runner.run(
		"What is the best movie in Korea? Have Korea movie ever won Oscar?"
	)
	assert isinstance(answer, str)
	assert bool(answer)


def test_runner_api_server(evaluator):
	project_dir = evaluator.project_dir
	evaluator.start_trial(os.path.join(resource_dir, "simple_mock.yaml"))
	runner = Runner.from_trial_folder(os.path.join(project_dir, "0"))

	client = runner.app.test_client()

	# Use the TestClient to make a request to the server
	response = client.post(
		"/run",
		json={
			"query": "What is the best movie in Korea? Have Korea movie ever won Oscar?",
			"result_column": "retrieved_contents",
		},
	)
	assert response.status_code == 200
	assert "retrieved_contents" in response.json
	retrieved_contents = response.json["retrieved_contents"]
	assert len(retrieved_contents) == 10
	assert isinstance(retrieved_contents, list)
	assert isinstance(retrieved_contents[0], str)



import gc
from copy import deepcopy
from typing import List, Tuple

import pandas as pd
import torch

from autorag.nodes.generator.base import BaseGenerator
from autorag.utils import result_to_dataframe
from autorag.utils.util import pop_params, to_list


class Vllm(BaseGenerator):
	def __init__(self, project_dir: str, llm: str, **kwargs):
		super().__init__(project_dir, llm, **kwargs)
		try:
			from vllm import SamplingParams, LLM
		except ImportError:
			raise ImportError(
				"Please install vllm library. You can install it by running `pip install vllm`."
			)

		model_from_kwargs = kwargs.pop("model", None)
		model = llm if model_from_kwargs is None else model_from_kwargs

		input_kwargs = deepcopy(kwargs)
		sampling_params_init_params = pop_params(
			SamplingParams.from_optional, input_kwargs
		)
		self.vllm_model = LLM(model, **input_kwargs)

		# delete not sampling param keys in the kwargs
		kwargs_keys = list(kwargs.keys())
		for key in kwargs_keys:
			if key not in sampling_params_init_params:
				kwargs.pop(key)

	def __del__(self):
		if torch.cuda.is_available():
			from vllm.distributed.parallel_state import (
				destroy_model_parallel,
			)

			destroy_model_parallel()
			del self.vllm_model
			gc.collect()
			torch.cuda.empty_cache()
			torch.distributed.destroy_process_group()
			torch.cuda.synchronize()
		else:
			del self.vllm_model

		super().__del__()

	@result_to_dataframe(["generated_texts", "generated_tokens", "generated_log_probs"])
	def pure(self, previous_result: pd.DataFrame, *args, **kwargs):
		prompts = self.cast_to_run(previous_result)
		return self._pure(prompts, **kwargs)

	def _pure(
		self, prompts: List[str], **kwargs
	) -> Tuple[List[str], List[List[int]], List[List[float]]]:
		"""
		Vllm module.
		It gets the VLLM instance and returns generated texts by the input prompt.
		You can set logprobs to get the log probs of the generated text.
		Default logprobs is 1.

		:param prompts: A list of prompts.
		:param kwargs: The extra parameters for generating the text.
		:return: A tuple of three elements.
		    The first element is a list of generated text.
		    The second element is a list of generated text's token ids.
		    The third element is a list of generated text's log probs.
		"""
		try:
			from vllm.outputs import RequestOutput
			from vllm.sequence import SampleLogprobs
			from vllm import SamplingParams
		except ImportError:
			raise ImportError(
				"Please install vllm library. You can install it by running `pip install vllm`."
			)

		if "logprobs" not in kwargs:
			kwargs["logprobs"] = 1

		sampling_params = pop_params(SamplingParams.from_optional, kwargs)
		generate_params = SamplingParams(**sampling_params)
		results: List[RequestOutput] = self.vllm_model.generate(
			prompts, generate_params
		)
		generated_texts = list(map(lambda x: x.outputs[0].text, results))
		generated_token_ids = list(map(lambda x: x.outputs[0].token_ids, results))
		log_probs: List[SampleLogprobs] = list(
			map(lambda x: x.outputs[0].logprobs, results)
		)
		generated_log_probs = list(
			map(
				lambda x: list(map(lambda y: y[0][y[1]].logprob, zip(x[0], x[1]))),
				zip(log_probs, generated_token_ids),
			)
		)
		return (
			to_list(generated_texts),
			to_list(generated_token_ids),
			to_list(generated_log_probs),
		)



import os
import tempfile
from unittest.mock import patch

import pandas as pd
import pytest
from llama_index.embeddings.openai import OpenAIEmbedding

from autorag.nodes.passageaugmenter import PrevNextPassageAugmenter
from autorag.nodes.passageaugmenter.run import run_passage_augmenter_node
from autorag.utils.util import load_summary_file
from tests.autorag.nodes.passageaugmenter.test_base_passage_augmenter import (
	qa_data,
	corpus_data,
	previous_result,
)
from tests.mock import mock_get_text_embedding_batch


@pytest.fixture
def node_line_dir():
	with tempfile.TemporaryDirectory(ignore_cleanup_errors=True) as project_dir:
		data_dir = os.path.join(project_dir, "data")
		os.makedirs(data_dir)
		qa_data.to_parquet(os.path.join(data_dir, "qa.parquet"), index=False)
		corpus_data.to_parquet(os.path.join(data_dir, "corpus.parquet"), index=False)
		trial_dir = os.path.join(project_dir, "0")
		os.makedirs(trial_dir)
		node_line_dir = os.path.join(trial_dir, "node_line_1")
		os.makedirs(node_line_dir)
		yield node_line_dir


@patch.object(
	OpenAIEmbedding,
	"get_text_embedding_batch",
	mock_get_text_embedding_batch,
)
def test_run_passage_augmenter_node(node_line_dir):
	modules = [PrevNextPassageAugmenter]
	module_params = [{"top_k": 2, "num_passages": 1}]
	strategies = {
		"metrics": ["retrieval_f1", "retrieval_recall"],
		"strategy": "rank",
	}
	best_result = run_passage_augmenter_node(
		modules, module_params, previous_result, node_line_dir, strategies
	)
	assert os.path.exists(os.path.join(node_line_dir, "passage_augmenter"))
	assert set(best_result.columns) == {
		"qid",
		"query",
		"retrieval_gt",
		"generation_gt",
		"retrieved_contents",
		"retrieved_ids",
		"retrieve_scores",
		"retrieval_f1",
		"retrieval_recall",
		"passage_augmenter_retrieval_f1",
		"passage_augmenter_retrieval_recall",
	}
	# test summary feature
	summary_path = os.path.join(node_line_dir, "passage_augmenter", "summary.csv")
	assert os.path.exists(summary_path)
	result_path = os.path.join(node_line_dir, "passage_augmenter", "0.parquet")
	assert os.path.exists(result_path)
	result_df = pd.read_parquet(result_path)
	summary_df = load_summary_file(summary_path)
	assert set(summary_df.columns) == {
		"filename",
		"passage_augmenter_retrieval_f1",
		"passage_augmenter_retrieval_recall",
		"module_name",
		"module_params",
		"execution_time",
		"is_best",
	}
	assert len(summary_df) == 1
	assert summary_df["filename"][0] == "0.parquet"
	assert summary_df["passage_augmenter_retrieval_f1"][0] == pytest.approx(
		result_df["retrieval_f1"].mean()
	)
	assert summary_df["passage_augmenter_retrieval_recall"][0] == pytest.approx(
		result_df["retrieval_recall"].mean()
	)
	assert summary_df["module_name"][0] == "PrevNextPassageAugmenter"
	assert summary_df["module_params"][0] == {"top_k": 2, "num_passages": 1}
	assert summary_df["execution_time"][0] > 0
	# test the best file is saved properly
	best_path = summary_df[summary_df["is_best"]]["filename"].values[0]
	assert os.path.exists(
		os.path.join(node_line_dir, "passage_augmenter", f"best_{best_path}")
	)



from typing import List
from unittest.mock import patch

from autorag.data.parse.llamaparse import llama_parse as llamaparse

from tests.autorag.data.parse.test_parse_base import (
	single_pdf_path_list,
	multiple_pdf_data_list,
	korean_text_glob,
	eng_text_glob,
	check_parse_result,
)

from llama_index.core.schema import Document
import llama_parse


async def mock_llama_parse_aload_data(*args, **kwargs) -> List[Document]:
	return [Document(id_="test_id", text="I love AutoRAG", metadata={})]


@patch.object(llama_parse.base.LlamaParse, "aload_data", mock_llama_parse_aload_data)
def test_llama_parse_single_pdf():
	llama_parse_original = llamaparse.__wrapped__
	texts, path, pages = llama_parse_original(
		single_pdf_path_list, url="mock_url", api_key="mock_api_key"
	)
	check_parse_result(texts, path, pages, "single_pdf", "llama")
	assert texts == ["I love AutoRAG"]


@patch.object(llama_parse.base.LlamaParse, "aload_data", mock_llama_parse_aload_data)
def test_llama_parse_single_pdf_node():
	result_df = llamaparse(korean_text_glob, url="mock_url", api_key="mock_api_key")
	check_parse_result(
		result_df["texts"].tolist(),
		result_df["path"].tolist(),
		result_df["page"].tolist(),
		"single_pdf",
		"llama",
	)
	assert result_df["texts"].tolist() == ["I love AutoRAG"]


@patch.object(llama_parse.base.LlamaParse, "aload_data", mock_llama_parse_aload_data)
def test_llama_parse_multiple_pdf():
	llama_parse_original = llamaparse.__wrapped__
	texts, path, pages = llama_parse_original(
		multiple_pdf_data_list, url="mock_url", api_key="mock_api_key"
	)
	check_parse_result(texts, path, pages, "multiple_pdf", "llama")
	assert texts == ["I love AutoRAG", "I love AutoRAG"]


@patch.object(llama_parse.base.LlamaParse, "aload_data", mock_llama_parse_aload_data)
def test_llama_parse_multiple_pdf_node():
	result_df = llamaparse(eng_text_glob, url="mock_url", api_key="mock_api_key")
	check_parse_result(
		result_df["texts"].tolist(),
		result_df["path"].tolist(),
		result_df["page"].tolist(),
		"multiple_pdf",
		"llama",
	)
	assert result_df["texts"].tolist() == ["I love AutoRAG", "I love AutoRAG"]



import itertools
import os
import pathlib

import pandas as pd

from autorag.data.chunk.base import make_metadata_list

root_dir = pathlib.PurePath(
	os.path.dirname(os.path.realpath(__file__))
).parent.parent.parent
resource_dir = os.path.join(root_dir, "resources")
data_dir = os.path.join(resource_dir, "chunk_data")

base_texts = [
	"The Kia Tigers lost the Korean Series this year and failed to win the championship. jeffrey went to "
	"gwangju to the Korean Series, but they lost there. I love this story.",
	"minsing's Real Madrid were crushed by Ulsan Hyundai of Korea's BOBB. minsing's Man United beat "
	"estdside_gunn's Chelsea. estdside_gunn always loses. I love this story.",
]
base_metadata = [
	{"path": "kia/lose/jeffrey_tigers_sad_story.pdf"},
	{"path": "chelsea/lose/gunn_chelsea_sad_story.pdf"},
]

parsed_result = pd.read_parquet(os.path.join(data_dir, "sample_parsed.parquet"))

character_expect_texts = list(
	itertools.chain.from_iterable(list(map(lambda x: x.split(". "), base_texts)))
)
expect_texts = {
	"token": [
		"The Kia Tigers lost the Korean Series this year and failed to win the "
		"championship. jeffrey went to gwangju to the Korean Series, but they lost",
		"there. I love this story.",
		"minsing's Real Madrid were crushed by Ulsan Hyundai of Korea's BOBB. "
		"minsing's Man United beat estdside_gunn's Chelsea.",
		"estdside_gunn always loses. I love this story.",
	],
	"token_ko": [
		"파일 제목: jeffrey_tigers_sad_story.pdf\n"
		" 내용: The Kia Tigers lost the Korean Series this year and failed to win the "
		"championship. jeffrey went to gwangju to the Korean Series, but they lost",
		"파일 제목: jeffrey_tigers_sad_story.pdf\n 내용: there. I love this story.",
		"파일 제목: gunn_chelsea_sad_story.pdf\n"
		" 내용: minsing's Real Madrid were crushed by Ulsan Hyundai of Korea's BOBB. "
		"minsing's Man United beat estdside_gunn's Chelsea.",
		"파일 제목: gunn_chelsea_sad_story.pdf\n"
		" 내용: estdside_gunn always loses. I love this story.",
	],
	"token_eng": [
		"file_name: jeffrey_tigers_sad_story.pdf\n"
		" contents: The Kia Tigers lost the Korean Series this year and failed to win "
		"the championship. jeffrey went to gwangju to the Korean Series, but they "
		"lost",
		"file_name: jeffrey_tigers_sad_story.pdf\n"
		" contents: there. I love this story.",
		"file_name: gunn_chelsea_sad_story.pdf\n"
		" contents: minsing's Real Madrid were crushed by Ulsan Hyundai of Korea's "
		"BOBB. minsing's Man United beat estdside_gunn's Chelsea.",
		"file_name: gunn_chelsea_sad_story.pdf\n"
		" contents: estdside_gunn always loses. I love this story.",
	],
	"character_ko": [
		f"파일 제목: jeffrey_tigers_sad_story.pdf\n 내용: {character_expect_texts[0]}",
		f"파일 제목: jeffrey_tigers_sad_story.pdf\n 내용: {character_expect_texts[1]}",
		f"파일 제목: jeffrey_tigers_sad_story.pdf\n 내용: {character_expect_texts[2]}",
		f"파일 제목: gunn_chelsea_sad_story.pdf\n 내용: {character_expect_texts[3]}",
		f"파일 제목: gunn_chelsea_sad_story.pdf\n 내용: {character_expect_texts[4]}",
		f"파일 제목: gunn_chelsea_sad_story.pdf\n 내용: {character_expect_texts[5]}",
		f"파일 제목: gunn_chelsea_sad_story.pdf\n 내용: {character_expect_texts[6]}",
	],
	"character_eng": [
		f"file_name: jeffrey_tigers_sad_story.pdf\n contents: {character_expect_texts[0]}",
		f"file_name: jeffrey_tigers_sad_story.pdf\n contents: {character_expect_texts[1]}",
		f"file_name: jeffrey_tigers_sad_story.pdf\n contents: {character_expect_texts[2]}",
		f"file_name: gunn_chelsea_sad_story.pdf\n contents: {character_expect_texts[3]}",
		f"file_name: gunn_chelsea_sad_story.pdf\n contents: {character_expect_texts[4]}",
		f"file_name: gunn_chelsea_sad_story.pdf\n contents: {character_expect_texts[5]}",
		f"file_name: gunn_chelsea_sad_story.pdf\n contents: {character_expect_texts[6]}",
	],
	"overlap": [
		"The Kia Tigers lost the Korean Series this year and failed to win the championship. jeffrey went to gwangju to the Korean Series, but they lost",
		"to the Korean Series, but they lost there. I love this story.",
		"minsing's Real Madrid were crushed by Ulsan Hyundai of Korea's BOBB. minsing's Man United beat estdside_gunn's Chelsea.",
		"United beat estdside_gunn's Chelsea. estdside_gunn always loses. I love this story.",
	],
}

expect_token_path = [
	base_metadata[0]["path"],
	base_metadata[0]["path"],
	base_metadata[1]["path"],
	base_metadata[1]["path"],
]

expect_character_path = [
	base_metadata[0]["path"],
	base_metadata[0]["path"],
	base_metadata[0]["path"],
	base_metadata[1]["path"],
	base_metadata[1]["path"],
	base_metadata[1]["path"],
	base_metadata[1]["path"],
]

expect_token_idx = [(0, 142), (144, 168), (0, 118), (120, 165)]
expect_overlap_idx = [(0, 142), (108, 168), (0, 118), (83, 165)]
expect_character_idx = [
	(0, 81),
	(84, 148),
	(151, 168),
	(0, 66),
	(69, 117),
	(120, 145),
	(148, 165),
]


def check_chunk_result(doc_id, contents, path, start_end_idx, metadata):
	params = [
		(doc_id, list, str),
		(contents, list, str),
		(path, list, str),
		(start_end_idx, list, tuple),
		(metadata, list, dict),
	]

	for param, outer_type, inner_type in params:
		assert isinstance(param, outer_type)
		assert isinstance(param[0], inner_type)


def check_chunk_result_node(result_df):
	check_chunk_result(
		result_df["doc_id"].tolist(),
		result_df["contents"].tolist(),
		result_df["path"].tolist(),
		result_df["start_end_idx"].tolist(),
		result_df["metadata"].tolist(),
	)


def test_make_metadata_list():
	parsed_df = pd.DataFrame(
		{
			"texts": base_texts,
			"page": [1, 1],
			"last_modified_datetime": ["24-09-03", "24-09-06"],
			"path": [
				"jax/jeffrey_tigers_sad_story.pdf",
				"siu/gunn_chelsea_sad_story.pdf",
			],
		}
	)
	meta_lst = make_metadata_list(parsed_df)
	assert meta_lst == [
		{
			"last_modified_datetime": "24-09-03",
			"page": 1,
			"path": "jax/jeffrey_tigers_sad_story.pdf",
		},
		{
			"last_modified_datetime": "24-09-06",
			"page": 1,
			"path": "siu/gunn_chelsea_sad_story.pdf",
		},
	]


def test_make_metadata_list_empty():
	parsed_df = pd.DataFrame({"texts": base_texts})
	meta_lst = make_metadata_list(parsed_df)
	assert meta_lst == [{}, {}]



import os
import pathlib

import pandas as pd

root_dir = pathlib.PurePath(
	os.path.dirname(os.path.realpath(__file__))
).parent.parent.parent
project_dir = os.path.join(root_dir, "resources", "sample_project")
qa_data = pd.read_parquet(os.path.join(project_dir, "data", "qa.parquet"))
corpus_data = pd.read_parquet(os.path.join(project_dir, "data", "corpus.parquet"))
doc_id_list = corpus_data["doc_id"].tolist()
previous_result = pd.DataFrame(
	{
		"qid": [1, 2],
		"query": [
			"What is the capital of France?",
			"How many members are in Newjeans?",
		],
		"retrieved_ids": [
			[doc_id_list[1], doc_id_list[3]],
			[doc_id_list[5], doc_id_list[7]],
		],
		"retrieved_contents": [
			[
				"Paris is the capital of France.",
				"Paris is one of the capital from France. Isn't it?",
			],
			["Newjeans has 5 members.", "Danielle is one of the members of Newjeans."],
		],
		"retrieve_scores": [[0.8, 0.1], [0.2, 0.1]],
		"retrieval_gt": [1, 1],
		"generation_gt": ["answer", "answer"],
		"retrieval_f1": [0.4, 1.0],
		"retrieval_recall": [1.0, 0.3],
	}
)
ids_list = [[doc_id_list[1]], [doc_id_list[3]], [doc_id_list[0], doc_id_list[29]]]



from typing import List, Tuple

import numpy as np
import pandas as pd
import torch
from tqdm import tqdm
from transformers import AutoModel, AutoTokenizer

from autorag.nodes.passagereranker.base import BasePassageReranker
from autorag.utils.util import (
	flatten_apply,
	sort_by_scores,
	select_top_k,
	pop_params,
	result_to_dataframe,
)


class ColbertReranker(BasePassageReranker):
	def __init__(
		self,
		project_dir: str,
		model_name: str = "colbert-ir/colbertv2.0",
		*args,
		**kwargs,
	):
		"""
		Initialize a colbert rerank model for reranking.

		:param project_dir: The project directory
		:param model_name: The model name for Colbert rerank.
			You can choose a colbert model for reranking.
			The default is "colbert-ir/colbertv2.0".
		:param kwargs: Extra parameter for the model.
		"""
		super().__init__(project_dir)
		self.device = "cuda" if torch.cuda.is_available() else "cpu"
		model_params = pop_params(AutoModel.from_pretrained, kwargs)
		self.model = AutoModel.from_pretrained(model_name, **model_params).to(
			self.device
		)
		self.tokenizer = AutoTokenizer.from_pretrained(model_name)

	def __del__(self):
		del self.model
		if torch.cuda.is_available():
			torch.cuda.empty_cache()
		super().__del__()

	@result_to_dataframe(["retrieved_contents", "retrieved_ids", "retrieve_scores"])
	def pure(self, previous_result: pd.DataFrame, *args, **kwargs):
		queries, contents, _, ids = self.cast_to_run(previous_result)
		top_k = kwargs.pop("top_k")
		batch = kwargs.pop("batch", 64)
		return self._pure(queries, contents, ids, top_k, batch)

	def _pure(
		self,
		queries: List[str],
		contents_list: List[List[str]],
		ids_list: List[List[str]],
		top_k: int,
		batch: int = 64,
	) -> Tuple[List[List[str]], List[List[str]], List[List[float]]]:
		"""
		Rerank a list of contents with Colbert rerank models.
		You can get more information about a Colbert model at https://huggingface.co/colbert-ir/colbertv2.0.
		It uses BERT-based model, so recommend using CUDA gpu for faster reranking.

		:param queries: The list of queries to use for reranking
		:param contents_list: The list of lists of contents to rerank
		:param ids_list: The list of lists of ids retrieved from the initial ranking
		:param top_k: The number of passages to be retrieved
		:param batch: The number of queries to be processed in a batch
			Default is 64.

		:return: Tuple of lists containing the reranked contents, ids, and scores
		"""

		# get query and content embeddings
		query_embedding_list = get_colbert_embedding_batch(
			queries, self.model, self.tokenizer, batch
		)
		content_embedding_list = flatten_apply(
			get_colbert_embedding_batch,
			contents_list,
			model=self.model,
			tokenizer=self.tokenizer,
			batch_size=batch,
		)
		df = pd.DataFrame(
			{
				"ids": ids_list,
				"query_embedding": query_embedding_list,
				"contents": contents_list,
				"content_embedding": content_embedding_list,
			}
		)
		temp_df = df.explode("content_embedding")
		temp_df["score"] = temp_df.apply(
			lambda x: get_colbert_score(x["query_embedding"], x["content_embedding"]),
			axis=1,
		)
		df["scores"] = (
			temp_df.groupby(level=0, sort=False)["score"].apply(list).tolist()
		)
		df[["contents", "ids", "scores"]] = df.apply(
			sort_by_scores, axis=1, result_type="expand"
		)
		results = select_top_k(df, ["contents", "ids", "scores"], top_k)

		return (
			results["contents"].tolist(),
			results["ids"].tolist(),
			results["scores"].tolist(),
		)


def get_colbert_embedding_batch(
	input_strings: List[str], model, tokenizer, batch_size: int
) -> List[np.array]:
	encoding = tokenizer(
		input_strings,
		return_tensors="pt",
		padding=True,
		truncation=True,
		max_length=model.config.max_position_embeddings,
	)

	input_batches = slice_tokenizer_result(encoding, batch_size)
	result_embedding = []
	with torch.no_grad():
		for encoding_batch in tqdm(input_batches):
			result_embedding.append(model(**encoding_batch).last_hidden_state)
	total_tensor = torch.cat(
		result_embedding, dim=0
	)  # shape [batch_size, token_length, embedding_dim]
	tensor_results = list(total_tensor.chunk(total_tensor.size()[0]))

	if torch.cuda.is_available():
		return list(map(lambda x: x.detach().cpu().numpy(), tensor_results))
	else:
		return list(map(lambda x: x.detach().numpy(), tensor_results))


def slice_tokenizer_result(tokenizer_output, batch_size):
	input_ids_batches = slice_tensor(tokenizer_output["input_ids"], batch_size)
	attention_mask_batches = slice_tensor(
		tokenizer_output["attention_mask"], batch_size
	)
	token_type_ids_batches = slice_tensor(
		tokenizer_output.get("token_type_ids", None), batch_size
	)
	return [
		{
			"input_ids": input_ids,
			"attention_mask": attention_mask,
			"token_type_ids": token_type_ids,
		}
		for input_ids, attention_mask, token_type_ids in zip(
			input_ids_batches, attention_mask_batches, token_type_ids_batches
		)
	]


def slice_tensor(input_tensor, batch_size):
	# Calculate the number of full batches
	num_full_batches = input_tensor.size(0) // batch_size

	# Slice the tensor into batches
	tensor_list = [
		input_tensor[i * batch_size : (i + 1) * batch_size]
		for i in range(num_full_batches)
	]

	# Handle the last batch if it's smaller than batch_size
	remainder = input_tensor.size(0) % batch_size
	if remainder:
		tensor_list.append(input_tensor[-remainder:])

	device = "cuda" if torch.cuda.is_available() else "cpu"
	tensor_list = list(map(lambda x: x.to(device), tensor_list))

	return tensor_list


def get_colbert_score(query_embedding: np.array, content_embedding: np.array) -> float:
	if query_embedding.ndim == 3 and content_embedding.ndim == 3:
		query_embedding = query_embedding.reshape(-1, query_embedding.shape[-1])
		content_embedding = content_embedding.reshape(-1, content_embedding.shape[-1])

	sim_matrix = np.dot(query_embedding, content_embedding.T) / (
		np.linalg.norm(query_embedding, axis=1)[:, np.newaxis]
		* np.linalg.norm(content_embedding, axis=1)
	)
	max_sim_scores = np.max(sim_matrix, axis=1)
	return float(np.mean(max_sim_scores))



import pandas as pd
import pytest

from autorag import generator_models
from autorag.nodes.generator import LlamaIndexLLM
from tests.autorag.nodes.generator.test_generator_base import (
	prompts,
	check_generated_texts,
	check_generated_tokens,
	check_generated_log_probs,
)
from tests.mock import MockLLM


@pytest.fixture
def llama_index_llm_instance():
	generator_models["mock"] = MockLLM
	return LlamaIndexLLM(project_dir=".", llm="mock", temperature=0.5, top_p=0.9)


def test_llama_index_llm(llama_index_llm_instance):
	answers, tokens, log_probs = llama_index_llm_instance._pure(prompts)
	check_generated_texts(answers)
	check_generated_tokens(tokens)
	check_generated_log_probs(log_probs)
	assert all(
		all(log_prob == 0.5 for log_prob in log_prob_list)
		for log_prob_list in log_probs
	)
	assert all(len(tokens[i]) == len(log_probs[i]) for i in range(len(tokens)))


def test_llama_index_llm_node():
	generator_models["mock"] = MockLLM
	previous_result = pd.DataFrame(
		{"prompts": prompts, "qid": ["id-1", "id-2", "id-3"]}
	)
	result_df = LlamaIndexLLM.run_evaluator(
		project_dir=".",
		previous_result=previous_result,
		llm="mock",
		temperature=0.5,
		top_p=0.9,
	)
	check_generated_texts(result_df["generated_texts"].tolist())
	check_generated_tokens(result_df["generated_tokens"].tolist())
	check_generated_log_probs(result_df["generated_log_probs"].tolist())



from autorag.nodes.passageaugmenter import PassPassageAugmenter

from tests.autorag.nodes.passageaugmenter.test_base_passage_augmenter import (
	project_dir,
	previous_result,
)


def test_pass_passage_augmenter():
	result_df = PassPassageAugmenter.run_evaluator(
		project_dir=project_dir, previous_result=previous_result, top_k=2
	)
	assert all(
		[
			column_name in result_df.columns
			for column_name in [
				"retrieved_contents",
				"retrieved_ids",
				"retrieve_scores",
			]
		]
	)
	assert (
		result_df["retrieved_contents"].tolist()
		== previous_result["retrieved_contents"].tolist()
	)
	assert (
		result_df["retrieved_ids"].tolist() == previous_result["retrieved_ids"].tolist()
	)
	assert (
		result_df["retrieve_scores"].tolist()
		== previous_result["retrieve_scores"].tolist()
	)



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = "AutoRAG"
copyright = "2024, Marker-Inc"
author = "Marker-Inc"

with open("../../autorag/VERSION") as f:
	version = f.read().strip()

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
	"sphinx.ext.autodoc",
	"sphinx.ext.extlinks",
	"sphinx.ext.intersphinx",
	"sphinx.ext.mathjax",
	"sphinx.ext.todo",
	"sphinx.ext.viewcode",
	"myst_parser",
	"sphinx_copybutton",
	"sphinx_design",
	"sphinx_inline_tabs",
	"sphinxcontrib.googleanalytics",
	"sphinx_sitemap",
]
source_suffix = {
	".rst": "restructuredtext",
	".md": "markdown",
}

templates_path = ["_templates"]
exclude_patterns = []

googleanalytics_id = "G-T8ZKQM6RT9"

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "furo"
html_static_path = ["_static"]
html_baseurl = "https://docs.auto-rag.com/"



import pandas as pd

from autorag.nodes.passagecompressor import PassCompressor

queries = [
	"What is the capital of France?",
	"What is the meaning of life?",
]
retrieved_contents = [
	[
		"Paris is the capital of France.",
		"France is a country in Europe.",
		"France is a member of the EU.",
	],
	[
		"The meaning of life is 42.",
		"The meaning of life is to be happy.",
		"The meaning of life is to be kind.",
	],
]


def test_pass_compressor():
	df = pd.DataFrame(
		{
			"query": queries,
			"retrieved_contents": retrieved_contents,
			"retrieved_ids": [["id-1", "id-2", "id-3"], ["id-4", "id-5", "id-6"]],
			"retrieve_scores": [[0.1, 0.2, 0.3], [0.4, 0.5, 0.6]],
		}
	)
	result_df = PassCompressor.run_evaluator("project_dir", df)
	assert result_df["retrieved_contents"].tolist() == retrieved_contents



import os
import pathlib
import subprocess

import pytest


from autorag.deploy import Runner

root_dir = pathlib.PurePath(os.path.dirname(os.path.realpath(__file__))).parent
resource_dir = os.path.join(root_dir, "resources")


@pytest.mark.skip(
	reason="The only way to test it is to run it directly on the web and see if it works."
	"If you want to test it, clear this line and run it."
)
def test_web_cli_yaml_project_dir():
	subprocess.run(
		[
			"autorag",
			"run_web",
			"--yaml_path",
			"../../sample_config/best.yaml",
			"--project_dir",
			"../resources/result_project",
		]
	)


@pytest.mark.skip(
	reason="The only way to test it is to run it directly on the web and see if it works."
	"If you want to test it, clear this line and run it."
)
def test_web_cli_trial():
	subprocess.run(
		["autorag", "run_web", "--trial_path", "../resources/result_project/0"]
	)


@pytest.mark.skip(
	reason="The only way to test it is to run it directly on the web and see if it works."
	"If you want to test it, clear this line and run it."
)
def test_web_library():
	runner = Runner.from_trial_folder(os.path.join(resource_dir, "result_project", "0"))
	runner.run_web()



import logging
import os
import pathlib
from copy import deepcopy
from typing import List, Callable, Dict, Tuple, Union

import numpy as np
import pandas as pd
from tqdm import tqdm

from autorag.evaluation import evaluate_retrieval
from autorag.schema.metricinput import MetricInput
from autorag.strategy import measure_speed, filter_by_threshold, select_best
from autorag.support import get_support_modules
from autorag.utils.util import get_best_row, to_list, apply_recursive

logger = logging.getLogger("AutoRAG")

semantic_module_names = ["vectordb", "VectorDB"]
lexical_module_names = ["bm25", "BM25"]
hybrid_module_names = ["hybrid_rrf", "hybrid_cc", "HybridCC", "HybridRRF"]


def run_retrieval_node(
	modules: List,
	module_params: List[Dict],
	previous_result: pd.DataFrame,
	node_line_dir: str,
	strategies: Dict,
) -> pd.DataFrame:
	"""
	Run evaluation and select the best module among retrieval node results.

	:param modules: Retrieval modules to run.
	:param module_params: Retrieval module parameters.
	:param previous_result: Previous result dataframe.
	    Could be query expansion's best result or qa data.
	:param node_line_dir: This node line's directory.
	:param strategies: Strategies for retrieval node.
	:return: The best result dataframe.
	    It contains previous result columns and retrieval node's result columns.
	"""
	if not os.path.exists(node_line_dir):
		os.makedirs(node_line_dir)
	project_dir = pathlib.PurePath(node_line_dir).parent.parent
	qa_df = pd.read_parquet(
		os.path.join(project_dir, "data", "qa.parquet"), engine="pyarrow"
	)
	retrieval_gt = qa_df["retrieval_gt"].tolist()
	retrieval_gt = apply_recursive(lambda x: str(x), to_list(retrieval_gt))
	# make rows to metric_inputs
	metric_inputs = [
		MetricInput(retrieval_gt=ret_gt, query=query, generation_gt=gen_gt)
		for ret_gt, query, gen_gt in zip(
			retrieval_gt, qa_df["query"].tolist(), qa_df["generation_gt"].tolist()
		)
	]

	save_dir = os.path.join(node_line_dir, "retrieval")  # node name
	if not os.path.exists(save_dir):
		os.makedirs(save_dir)

	def run(input_modules, input_module_params) -> Tuple[List[pd.DataFrame], List]:
		"""
		Run input modules and parameters.

		:param input_modules: Input modules
		:param input_module_params: Input module parameters
		:return: First, it returns list of result dataframe.
		Second, it returns list of execution times.
		"""
		result, execution_times = zip(
			*map(
				lambda task: measure_speed(
					task[0].run_evaluator,
					project_dir=project_dir,
					previous_result=previous_result,
					**task[1],
				),
				zip(input_modules, input_module_params),
			)
		)
		average_times = list(map(lambda x: x / len(result[0]), execution_times))

		# run metrics before filtering
		if strategies.get("metrics") is None:
			raise ValueError("You must at least one metrics for retrieval evaluation.")
		result = list(
			map(
				lambda x: evaluate_retrieval_node(
					x,
					metric_inputs,
					strategies.get("metrics"),
				),
				result,
			)
		)

		return result, average_times

	def save_and_summary(
		input_modules,
		input_module_params,
		result_list,
		execution_time_list,
		filename_start: int,
	):
		"""
		Save the result and make summary file

		:param input_modules: Input modules
		:param input_module_params: Input module parameters
		:param result_list: Result list
		:param execution_time_list: Execution times
		:param filename_start: The first filename to use
		:return: First, it returns list of result dataframe.
		Second, it returns list of execution times.
		"""

		# save results to folder
		filepaths = list(
			map(
				lambda x: os.path.join(save_dir, f"{x}.parquet"),
				range(filename_start, filename_start + len(input_modules)),
			)
		)
		list(
			map(
				lambda x: x[0].to_parquet(x[1], index=False),
				zip(result_list, filepaths),
			)
		)  # execute save to parquet
		filename_list = list(map(lambda x: os.path.basename(x), filepaths))

		summary_df = pd.DataFrame(
			{
				"filename": filename_list,
				"module_name": list(map(lambda module: module.__name__, input_modules)),
				"module_params": input_module_params,
				"execution_time": execution_time_list,
				**{
					metric: list(map(lambda result: result[metric].mean(), result_list))
					for metric in strategies.get("metrics")
				},
			}
		)
		summary_df.to_csv(os.path.join(save_dir, "summary.csv"), index=False)
		return summary_df

	def find_best(results, average_times, filenames):
		# filter by strategies
		if strategies.get("speed_threshold") is not None:
			results, filenames = filter_by_threshold(
				results, average_times, strategies["speed_threshold"], filenames
			)
		selected_result, selected_filename = select_best(
			results,
			strategies.get("metrics"),
			filenames,
			strategies.get("strategy", "mean"),
		)
		return selected_result, selected_filename

	filename_first = 0
	# run semantic modules
	logger.info("Running retrieval node - semantic retrieval module...")
	if any([module.__name__ in semantic_module_names for module in modules]):
		semantic_modules, semantic_module_params = zip(
			*filter(
				lambda x: x[0].__name__ in semantic_module_names,
				zip(modules, module_params),
			)
		)
		semantic_results, semantic_times = run(semantic_modules, semantic_module_params)
		semantic_summary_df = save_and_summary(
			semantic_modules,
			semantic_module_params,
			semantic_results,
			semantic_times,
			filename_first,
		)
		semantic_selected_result, semantic_selected_filename = find_best(
			semantic_results, semantic_times, semantic_summary_df["filename"].tolist()
		)
		semantic_summary_df["is_best"] = (
			semantic_summary_df["filename"] == semantic_selected_filename
		)
		filename_first += len(semantic_modules)
	else:
		(
			semantic_selected_filename,
			semantic_summary_df,
			semantic_results,
			semantic_times,
		) = None, pd.DataFrame(), [], []
	# run lexical modules
	logger.info("Running retrieval node - lexical retrieval module...")
	if any([module.__name__ in lexical_module_names for module in modules]):
		lexical_modules, lexical_module_params = zip(
			*filter(
				lambda x: x[0].__name__ in lexical_module_names,
				zip(modules, module_params),
			)
		)
		lexical_results, lexical_times = run(lexical_modules, lexical_module_params)
		lexical_summary_df = save_and_summary(
			lexical_modules,
			lexical_module_params,
			lexical_results,
			lexical_times,
			filename_first,
		)
		lexical_selected_result, lexical_selected_filename = find_best(
			lexical_results, lexical_times, lexical_summary_df["filename"].tolist()
		)
		lexical_summary_df["is_best"] = (
			lexical_summary_df["filename"] == lexical_selected_filename
		)
		filename_first += len(lexical_modules)
	else:
		(
			lexical_selected_filename,
			lexical_summary_df,
			lexical_results,
			lexical_times,
		) = None, pd.DataFrame(), [], []

	logger.info("Running retrieval node - hybrid retrieval module...")
	# Next, run hybrid retrieval
	if any([module.__name__ in hybrid_module_names for module in modules]):
		hybrid_modules, hybrid_module_params = zip(
			*filter(
				lambda x: x[0].__name__ in hybrid_module_names,
				zip(modules, module_params),
			)
		)
		if all(
			["target_module_params" in x for x in hybrid_module_params]
		):  # for Runner.run
			# If target_module_params are already given, run hybrid retrieval directly
			hybrid_results, hybrid_times = run(hybrid_modules, hybrid_module_params)
			hybrid_summary_df = save_and_summary(
				hybrid_modules,
				hybrid_module_params,
				hybrid_results,
				hybrid_times,
				filename_first,
			)
			filename_first += len(hybrid_modules)
		else:  # for Evaluator
			# get id and score
			ids_scores = get_ids_and_scores(
				save_dir,
				[semantic_selected_filename, lexical_selected_filename],
				semantic_summary_df,
				lexical_summary_df,
				previous_result,
			)
			hybrid_module_params = list(
				map(lambda x: {**x, **ids_scores}, hybrid_module_params)
			)

			# optimize each modules
			real_hybrid_times = [
				get_hybrid_execution_times(semantic_summary_df, lexical_summary_df)
			] * len(hybrid_module_params)
			hybrid_times = real_hybrid_times.copy()
			hybrid_results = []
			for module, module_param in zip(hybrid_modules, hybrid_module_params):
				module_result_df, module_best_weight = optimize_hybrid(
					module,
					module_param,
					strategies,
					metric_inputs,
					project_dir,
					previous_result,
				)
				module_param["weight"] = module_best_weight
				hybrid_results.append(module_result_df)

			hybrid_summary_df = save_and_summary(
				hybrid_modules,
				hybrid_module_params,
				hybrid_results,
				hybrid_times,
				filename_first,
			)
			filename_first += len(hybrid_modules)
			hybrid_summary_df["execution_time"] = hybrid_times
			best_semantic_summary_row = semantic_summary_df.loc[
				semantic_summary_df["is_best"]
			].iloc[0]
			best_lexical_summary_row = lexical_summary_df.loc[
				lexical_summary_df["is_best"]
			].iloc[0]
			target_modules = (
				best_semantic_summary_row["module_name"],
				best_lexical_summary_row["module_name"],
			)
			target_module_params = (
				best_semantic_summary_row["module_params"],
				best_lexical_summary_row["module_params"],
			)
			hybrid_summary_df = edit_summary_df_params(
				hybrid_summary_df, target_modules, target_module_params
			)
	else:
		if any([module.__name__ in hybrid_module_names for module in modules]):
			logger.warning(
				"You must at least one semantic module and lexical module for hybrid evaluation."
				"Passing hybrid module."
			)
		_, hybrid_summary_df, hybrid_results, hybrid_times = (
			None,
			pd.DataFrame(),
			[],
			[],
		)

	summary = pd.concat(
		[semantic_summary_df, lexical_summary_df, hybrid_summary_df], ignore_index=True
	)
	results = semantic_results + lexical_results + hybrid_results
	average_times = semantic_times + lexical_times + hybrid_times
	filenames = summary["filename"].tolist()

	# filter by strategies
	selected_result, selected_filename = find_best(results, average_times, filenames)
	best_result = pd.concat([previous_result, selected_result], axis=1)

	# add summary.csv 'is_best' column
	summary["is_best"] = summary["filename"] == selected_filename

	# save the result files
	best_result.to_parquet(
		os.path.join(
			save_dir, f"best_{os.path.splitext(selected_filename)[0]}.parquet"
		),
		index=False,
	)
	summary.to_csv(os.path.join(save_dir, "summary.csv"), index=False)
	return best_result


def evaluate_retrieval_node(
	result_df: pd.DataFrame,
	metric_inputs: List[MetricInput],
	metrics: Union[List[str], List[Dict]],
) -> pd.DataFrame:
	"""
	Evaluate retrieval node from retrieval node result dataframe.

	:param result_df: The result dataframe from a retrieval node.
	:param metric_inputs: List of metric input schema for AutoRAG.
	:param metrics: Metric list from input strategies.
	:return: Return result_df with metrics columns.
	    The columns will be 'retrieved_contents', 'retrieved_ids', 'retrieve_scores', and metric names.
	"""

	@evaluate_retrieval(
		metric_inputs=metric_inputs,
		metrics=metrics,
	)
	def evaluate_this_module(df: pd.DataFrame):
		return (
			df["retrieved_contents"].tolist(),
			df["retrieved_ids"].tolist(),
			df["retrieve_scores"].tolist(),
		)

	return evaluate_this_module(result_df)


def edit_summary_df_params(
	summary_df: pd.DataFrame, target_modules, target_module_params
) -> pd.DataFrame:
	def delete_ids_scores(x):
		del x["ids"]
		del x["scores"]
		return x

	summary_df["module_params"] = summary_df["module_params"].apply(delete_ids_scores)
	summary_df["new_params"] = [
		{"target_modules": target_modules, "target_module_params": target_module_params}
	] * len(summary_df)
	summary_df["module_params"] = summary_df.apply(
		lambda row: {**row["module_params"], **row["new_params"]}, axis=1
	)
	summary_df = summary_df.drop(columns=["new_params"])
	return summary_df


def get_ids_and_scores(
	node_dir: str,
	filenames: List[str],
	semantic_summary_df: pd.DataFrame,
	lexical_summary_df: pd.DataFrame,
	previous_result,
) -> Dict[str, Tuple[List[List[str]], List[List[float]]]]:
	project_dir = pathlib.PurePath(node_dir).parent.parent.parent
	best_results_df = list(
		map(
			lambda filename: pd.read_parquet(
				os.path.join(node_dir, filename), engine="pyarrow"
			),
			filenames,
		)
	)
	ids = tuple(
		map(lambda df: df["retrieved_ids"].apply(list).tolist(), best_results_df)
	)
	scores = tuple(
		map(lambda df: df["retrieve_scores"].apply(list).tolist(), best_results_df)
	)
	# search non-duplicate ids
	semantic_ids = deepcopy(ids[0])
	lexical_ids = deepcopy(ids[1])

	def get_non_duplicate_ids(target_ids, compare_ids) -> List[List[str]]:
		"""
		Get non-duplicate ids from target_ids and compare_ids.
		If you want to non-duplicate ids of semantic_ids, you have to put it at target_ids.
		"""
		result_ids = []
		assert len(target_ids) == len(compare_ids)
		for target_id_list, compare_id_list in zip(target_ids, compare_ids):
			query_duplicated = list(set(compare_id_list) - set(target_id_list))
			duplicate_list = query_duplicated if len(query_duplicated) != 0 else []
			result_ids.append(duplicate_list)
		return result_ids

	lexical_target_ids = get_non_duplicate_ids(lexical_ids, semantic_ids)
	semantic_target_ids = get_non_duplicate_ids(semantic_ids, lexical_ids)

	new_id_tuple = (
		[a + b for a, b in zip(semantic_ids, semantic_target_ids)],
		[a + b for a, b in zip(lexical_ids, lexical_target_ids)],
	)

	# search non-duplicate ids' scores
	new_semantic_scores = get_scores_by_ids(
		semantic_target_ids, semantic_summary_df, project_dir, previous_result
	)
	new_lexical_scores = get_scores_by_ids(
		lexical_target_ids, lexical_summary_df, project_dir, previous_result
	)

	new_score_tuple = (
		[a + b for a, b in zip(scores[0], new_semantic_scores)],
		[a + b for a, b in zip(scores[1], new_lexical_scores)],
	)
	return {
		"ids": new_id_tuple,
		"scores": new_score_tuple,
	}


def get_scores_by_ids(
	ids: List[List[str]], module_summary_df: pd.DataFrame, project_dir, previous_result
) -> List[List[float]]:
	module_name = get_best_row(module_summary_df)["module_name"]
	module_params = get_best_row(module_summary_df)["module_params"]
	module = get_support_modules(module_name)
	result_df = module.run_evaluator(
		project_dir=project_dir,
		previous_result=previous_result,
		ids=ids,
		**module_params,
	)
	return to_list(result_df["retrieve_scores"].tolist())


def find_unique_elems(list1: List[str], list2: List[str]) -> List[str]:
	return list(set(list1).symmetric_difference(set(list2)))


def get_hybrid_execution_times(lexical_summary, semantic_summary) -> float:
	lexical_execution_time = lexical_summary.loc[lexical_summary["is_best"]].iloc[0][
		"execution_time"
	]
	semantic_execution_time = semantic_summary.loc[semantic_summary["is_best"]].iloc[0][
		"execution_time"
	]
	return lexical_execution_time + semantic_execution_time


def optimize_hybrid(
	hybrid_module_func: Callable,
	hybrid_module_param: Dict,
	strategy: Dict,
	input_metrics: List[MetricInput],
	project_dir,
	previous_result,
):
	if (
		hybrid_module_func.__name__ == "HybridRRF"
		or hybrid_module_func.__name__ == "hybrid_rrf"
	):
		weight_range = hybrid_module_param.pop("weight_range", (4, 80))
		test_weight_size = weight_range[1] - weight_range[0] + 1
	elif (
		hybrid_module_func.__name__ == "HybridCC"
		or hybrid_module_func.__name__ == "hybrid_cc"
	):
		weight_range = hybrid_module_param.pop("weight_range", (0.0, 1.0))
		test_weight_size = hybrid_module_param.pop("test_weight_size", 101)
	else:
		raise ValueError("You must input hybrid module function at hybrid_module_func.")

	weight_candidates = np.linspace(
		weight_range[0], weight_range[1], test_weight_size
	).tolist()

	result_list = []
	for weight_value in tqdm(weight_candidates):
		result_df = hybrid_module_func.run_evaluator(
			project_dir=project_dir,
			previous_result=previous_result,
			weight=weight_value,
			**hybrid_module_param,
		)
		result_list.append(result_df)

		# evaluate here
	if strategy.get("metrics") is None:
		raise ValueError("You must at least one metrics for retrieval evaluation.")
	result_list = list(
		map(
			lambda x: evaluate_retrieval_node(
				x,
				input_metrics,
				strategy.get("metrics"),
			),
			result_list,
		)
	)

	# select best result
	best_result_df, best_weight = select_best(
		result_list,
		strategy.get("metrics"),
		metadatas=weight_candidates,
		strategy_name=strategy.get("strategy", "normalize_mean"),
	)
	return best_result_df, best_weight



import json
import logging
import os
import shutil
from datetime import datetime
from typing import Optional

from autorag.data.parse.run import run_parser
from autorag.data.utils.util import load_yaml, get_param_combinations

logger = logging.getLogger("AutoRAG")


class Parser:
	def __init__(self, data_path_glob: str, project_dir: Optional[str] = None):
		self.data_path_glob = data_path_glob
		self.project_dir = project_dir if project_dir is not None else os.getcwd()

	def start_parsing(self, yaml_path: str):
		trial_name = self.__get_new_trial_name()
		self.__make_trial_dir(trial_name)

		# copy yaml file to trial directory
		shutil.copy(
			yaml_path, os.path.join(self.project_dir, trial_name, "parse_config.yaml")
		)

		# load yaml file
		modules = load_yaml(yaml_path)

		input_modules, input_params = get_param_combinations(modules)

		logger.info("Parsing Start...")
		run_parser(
			modules=input_modules,
			module_params=input_params,
			data_path_glob=self.data_path_glob,
			trial_path=os.path.join(self.project_dir, trial_name),
		)
		logger.info("Parsing Done!")

	def __get_new_trial_name(self) -> str:
		trial_json_path = os.path.join(self.project_dir, "trial.json")
		if not os.path.exists(trial_json_path):
			return "0"
		with open(trial_json_path, "r") as f:
			trial_json = json.load(f)
		return str(int(trial_json[-1]["trial_name"]) + 1)

	def __make_trial_dir(self, trial_name: str):
		trial_json_path = os.path.join(self.project_dir, "trial.json")
		if os.path.exists(trial_json_path):
			with open(trial_json_path, "r") as f:
				trial_json = json.load(f)
		else:
			trial_json = []

		trial_json.append(
			{
				"trial_name": trial_name,
				"start_time": datetime.now().strftime("%Y-%m-%d %H:%M:%S"),
			}
		)
		os.makedirs(os.path.join(self.project_dir, trial_name))
		with open(trial_json_path, "w") as f:
			json.dump(trial_json, f, indent=4)



import abc
import functools
import logging
import os
from pathlib import Path
from typing import Union, Tuple, List

import pandas as pd

from autorag.schema.base import BaseModule
from autorag.utils import result_to_dataframe, validate_qa_dataset, fetch_contents

logger = logging.getLogger("AutoRAG")


class BasePassageFilter(BaseModule, metaclass=abc.ABCMeta):
	def __init__(self, project_dir: Union[str, Path], *args, **kwargs):
		logger.info(f"Initialize passage filter node - {self.__class__.__name__}")

	def __del__(self):
		logger.info(f"Prompt maker node - {self.__class__.__name__} module is deleted.")

	def cast_to_run(self, previous_result: pd.DataFrame, *args, **kwargs):
		logger.info(
			f"Running passage filter node - {self.__class__.__name__} module..."
		)
		validate_qa_dataset(previous_result)

		# find queries columns
		assert (
			"query" in previous_result.columns
		), "previous_result must have query column."
		queries = previous_result["query"].tolist()

		# find contents_list columns
		assert (
			"retrieved_contents" in previous_result.columns
		), "previous_result must have retrieved_contents column."
		contents = previous_result["retrieved_contents"].tolist()

		# find scores columns
		assert (
			"retrieve_scores" in previous_result.columns
		), "previous_result must have retrieve_scores column."
		scores = previous_result["retrieve_scores"].tolist()

		# find ids columns
		assert (
			"retrieved_ids" in previous_result.columns
		), "previous_result must have retrieved_ids column."
		ids = previous_result["retrieved_ids"].tolist()
		return queries, contents, scores, ids


# same with passage filter from now
def passage_filter_node(func):
	@functools.wraps(func)
	@result_to_dataframe(["retrieved_contents", "retrieved_ids", "retrieve_scores"])
	def wrapper(
		project_dir: Union[str, Path], previous_result: pd.DataFrame, *args, **kwargs
	) -> Tuple[List[List[str]], List[List[str]], List[List[float]]]:
		if func.__name__ == "recency_filter":
			corpus_df = pd.read_parquet(
				os.path.join(project_dir, "data", "corpus.parquet"), engine="pyarrow"
			)
			metadatas = fetch_contents(corpus_df, ids, column_name="metadata")
			times = [
				[time["last_modified_datetime"] for time in time_list]
				for time_list in metadatas
			]
			filtered_contents, filtered_ids, filtered_scores = func(
				contents_list=contents,
				scores_list=scores,
				ids_list=ids,
				time_list=times,
				*args,
				**kwargs,
			)
		else:
			filtered_contents, filtered_ids, filtered_scores = func(
				queries=queries,
				contents_list=contents,
				scores_list=scores,
				ids_list=ids,
				*args,
				**kwargs,
			)

		return filtered_contents, filtered_ids, filtered_scores

	return wrapper
